import	os
import	unittest
from	typing								import List
from	pygwarts.magical.philosophers_stone	import Transmutable
from	pygwarts.magical.time_turner.timers	import Callstamp
from	pygwarts.magical.time_turner.timers	import mostsec
from	pygwarts.magical.time_turner.timers	import mostsecfmt
from	pygwarts.magical.time_turner.timers	import DIRTtimer
from	pygwarts.irma.contrib				import LibraryContrib
from	pygwarts.tests.magical				import MagicalTestCase








class TimeTurnerTimersCases(MagicalTestCase):

	@classmethod
	def tearDownClass(cls):

		if	cls.clean_up:
			if	os.path.isfile(cls.TIMETURN_TIMERS_HANDLER): os.remove(cls.TIMETURN_TIMERS_HANDLER)

	@classmethod
	def setUpClass(cls):

		@mostsecfmt(positive=False)
		class Secfmter(Transmutable):
			def __call__(self, *args, **kwargs): return args[0]

		@mostsecfmt(positive=True)
		class PositiveSecfmter(Transmutable):
			def __call__(self, *args, **kwargs): return args[0]

		cls.secfmter = Secfmter()
		cls.posfmter = PositiveSecfmter()
		cls.make_loggy_file(cls, cls.TIMETURN_TIMERS_HANDLER)


		class Chrome(Transmutable):

			def __init__(self, *args, **kwargs):

				super().__init__(*args, **kwargs)
				self.RAM = 1000

			def __call__(self, operation :str, *, source :List[int]):

				self.loggy.info(f"Going to {operation}")
				self.consume_RAM(source)

			def consume_RAM(self, source :List[int]):

				source[0] *= 2
				self.loggy.info(f"Current RAM consumed {source[0]}")

		cls.browser = Chrome




	def test_Callstamp_loggy(self):

		callstamp = str(self.MAGICAL_ROOT /"callstamp.loggy")
		self.make_loggy_file(callstamp)


		@Callstamp
		class Stamper(Transmutable):

			def __call__(self):

				self.loggy.info("Making a stamp now")
				return	"Making a stamp now"

			class loggy(LibraryContrib):

				handler		= callstamp
				init_name	= "callstamp-1"


		loggy = []
		self.test_case = Stamper()
		current = self.test_case()
		self.assertEqual(current, "Making a stamp now")
		self.test_case.loggy.close()

		with open(callstamp) as case_loggy:
			for line in case_loggy : loggy.append(line.rstrip("\n")[16:])

		self.assertEqual(len(loggy),2)
		self.assertEqual(loggy[0], "@callstamp-1 INFO : Making a stamp now")
		self.assertTrue(loggy[1].startswith("@callstamp-1 INFO : Stamper finished in "))




		@Callstamp
		class Stamper(Transmutable):

			def __call__(self):

				self.loggy.info("Making a stamp now")
				return	"Making a stamp now"

			class loggy(LibraryContrib):

				handler		= callstamp
				init_name	= "callstamp-2"
				init_level	= 10


		loggy = []
		self.make_loggy_file(callstamp)
		self.test_case = Stamper()
		current = self.test_case()
		self.assertEqual(current, "Making a stamp now")
		self.test_case.loggy.close()

		with open(callstamp) as case_loggy:
			for line in case_loggy : loggy.append(line.rstrip("\n")[16:])

		self.assertEqual(len(loggy),4)
		self.assertTrue(loggy[0].startswith("@callstamp-2-Stamper DEBUG : Start point "))
		self.assertEqual(loggy[1], "@callstamp-2-Stamper INFO : Making a stamp now")
		self.assertTrue(loggy[2].startswith("@callstamp-2-Stamper DEBUG : End point "))
		self.assertTrue(loggy[3].startswith("@callstamp-2-Stamper INFO : Stamper finished in "))




		@Callstamp
		class Stamper(Transmutable):

			def __call__(self):

				self.loggy.info("Making a stamp now")
				return	"Making a stamp now"

			class loggy(LibraryContrib):

				handler			= callstamp
				init_name		= "callstamp-3"
				force_handover	= True


		loggy = []
		self.make_loggy_file(callstamp)
		self.test_case = Stamper()
		current = self.test_case()
		self.assertEqual(current, "Making a stamp now")
		self.test_case.loggy.close()

		with open(callstamp) as case_loggy:
			for line in case_loggy : loggy.append(line.rstrip("\n")[16:])

		self.assertEqual(len(loggy),2)
		self.assertEqual(loggy[0], "@callstamp-3-Stamper INFO : Making a stamp now")
		self.assertTrue(loggy[1].startswith("@callstamp-3-Stamper INFO : Stamper finished in "))
		if os.path.isfile(callstamp): os.remove(callstamp)








	def test_Callstamp_name_loggy(self):

		callstamp_n = str(self.MAGICAL_ROOT /"callstamp_n.loggy")
		self.make_loggy_file(callstamp_n)


		@Callstamp
		class Stamper(Transmutable):

			def __call__(self):

				self.loggy.info("Making a stamp now")
				return	"Making a stamp now"

			class shmoggy(LibraryContrib):

				handler		= callstamp_n
				init_name	= "callstamp_n-1"


		loggy = []
		self.test_case = Stamper()
		current = self.test_case()
		self.assertEqual(current, "Making a stamp now")
		self.test_case.loggy.close()

		with open(callstamp_n) as case_loggy:
			for line in case_loggy : loggy.append(line.rstrip("\n")[16:])

		self.assertEqual(len(loggy),2)
		self.assertEqual(loggy[0], "@callstamp_n-1 INFO : Making a stamp now")
		self.assertTrue(loggy[1].startswith("@callstamp_n-1 INFO : Stamper finished in "))




		@Callstamp
		class Stamper(Transmutable):

			def __call__(self):

				self.loggy.info("Making a stamp now")
				return	"Making a stamp now"

			class shmoggy(LibraryContrib):

				handler		= callstamp_n
				init_name	= "callstamp_n-2"
				init_level	= 10


		loggy = []
		self.make_loggy_file(callstamp_n)
		self.test_case = Stamper()
		current = self.test_case()
		self.assertEqual(current, "Making a stamp now")
		self.test_case.loggy.close()

		with open(callstamp_n) as case_loggy:
			for line in case_loggy : loggy.append(line.rstrip("\n")[16:])

		self.assertEqual(len(loggy),4)
		self.assertTrue(loggy[0].startswith("@callstamp_n-2-Stamper DEBUG : Start point "))
		self.assertEqual(loggy[1], "@callstamp_n-2-Stamper INFO : Making a stamp now")
		self.assertTrue(loggy[2].startswith("@callstamp_n-2-Stamper DEBUG : End point "))
		self.assertTrue(loggy[3].startswith("@callstamp_n-2-Stamper INFO : Stamper finished in "))




		@Callstamp
		class Stamper(Transmutable):

			def __call__(self):

				self.loggy.info("Making a stamp now")
				return	"Making a stamp now"

			class shmoggy(LibraryContrib):

				handler			= callstamp_n
				init_name		= "callstamp_n-3"
				force_handover	= True


		loggy = []
		self.make_loggy_file(callstamp_n)
		self.test_case = Stamper()
		current = self.test_case()
		self.assertEqual(current, "Making a stamp now")
		self.test_case.loggy.close()

		with open(callstamp_n) as case_loggy:
			for line in case_loggy : loggy.append(line.rstrip("\n")[16:])

		self.assertEqual(len(loggy),2)
		self.assertEqual(loggy[0], "@callstamp_n-3-Stamper INFO : Making a stamp now")
		self.assertTrue(loggy[1].startswith("@callstamp_n-3-Stamper INFO : Stamper finished in "))
		if os.path.isfile(callstamp_n): os.remove(callstamp_n)








	def test_Callstamp_no_loggy(self):

		@Callstamp
		class Stamper(Transmutable):
			def __call__(self):

				self.loggy.info("Making a stamp now")
				return	"Making a stamp now"


		self.test_case = Stamper()
		current = self.test_case()
		self.assertEqual(current, "Making a stamp now")








	def test_nocall_Callstamp(self):

		nocall = str(self.MAGICAL_ROOT /"nocall.loggy")
		self.make_loggy_file(nocall)


		@Callstamp
		class Stamper(Transmutable):

			def stamping(self): self.loggy.info("This is not the stamp it should be")
			class loggy(LibraryContrib):

				handler		= nocall
				init_name	= "nocall-1"


		loggy = []
		self.test_case = Stamper()
		self.test_case.stamping()
		self.test_case.loggy.close()

		with open(nocall) as case_loggy:
			for line in case_loggy : loggy.append(line.rstrip("\n")[16:])

		self.assertEqual(len(loggy),1)
		self.assertEqual(loggy[0], "@nocall-1 INFO : This is not the stamp it should be")


		self.assertRaisesRegex(

			TypeError,
			"Stamper object is not callable",
			self.test_case
		)




		@Callstamp
		class Stamper(Transmutable):
			def stamping(self): self.loggy.info("This is not the stamp it should be")


		self.test_case = Stamper()
		self.test_case.stamping()
		self.assertRaisesRegex(

			TypeError,
			"Stamper object is not callable",
			self.test_case
		)
		if	os.path.isfile(nocall): os.remove(nocall)
















	def test_mostsec_int_seconds(self):

		self.assertEqual(mostsec(0, positive=True), None)
		self.assertEqual(mostsec(0), "<1 ns")
		self.assertEqual(mostsec(1), "1 s")
		self.assertEqual(mostsec(59), "59 s")
		self.assertEqual(mostsec(60), "1 m")
		self.assertEqual(mostsec(61), "1 m 1 s")
		self.assertEqual(mostsec(119), "1 m 59 s")
		self.assertEqual(mostsec(120), "2 m")
		self.assertEqual(mostsec(120), "2 m")
		self.assertEqual(mostsec(3599), "59 m 59 s")
		self.assertEqual(mostsec(3600), "1 h")
		self.assertEqual(mostsec(3601), "1 h")
		self.assertEqual(mostsec(3659), "1 h")
		self.assertEqual(mostsec(3660), "1 h 1 m")
		self.assertEqual(mostsec(86399), "23 h 59 m")
		self.assertEqual(mostsec(86400), "1 d")
		self.assertEqual(mostsec(86401), "1 d")
		self.assertEqual(mostsec(86459), "1 d")
		self.assertEqual(mostsec(89999), "1 d")
		self.assertEqual(mostsec(90000), "1 d 1 h")
		self.assertEqual(mostsec(31535999), "364 d 23 h")
		self.assertEqual(mostsec(31536000), "365 d")
		self.assertEqual(mostsec(31539599), "365 d")
		self.assertEqual(mostsec(31539600), "365 d 1 h")

	def test_mostsec_int_seconds_negative(self):

		self.assertEqual(mostsec(-0, positive=True), None)
		self.assertEqual(mostsec(-0), "<1 ns")
		self.assertEqual(mostsec(-1), "1 s")
		self.assertEqual(mostsec(-59), "59 s")
		self.assertEqual(mostsec(-60), "1 m")
		self.assertEqual(mostsec(-61), "1 m 1 s")
		self.assertEqual(mostsec(-119), "1 m 59 s")
		self.assertEqual(mostsec(-120), "2 m")
		self.assertEqual(mostsec(-120), "2 m")
		self.assertEqual(mostsec(-3599), "59 m 59 s")
		self.assertEqual(mostsec(-3600), "1 h")
		self.assertEqual(mostsec(-3601), "1 h")
		self.assertEqual(mostsec(-3659), "1 h")
		self.assertEqual(mostsec(-3660), "1 h 1 m")
		self.assertEqual(mostsec(-86399), "23 h 59 m")
		self.assertEqual(mostsec(-86400), "1 d")
		self.assertEqual(mostsec(-86401), "1 d")
		self.assertEqual(mostsec(-86459), "1 d")
		self.assertEqual(mostsec(-89999), "1 d")
		self.assertEqual(mostsec(-90000), "1 d 1 h")
		self.assertEqual(mostsec(-31535999), "364 d 23 h")
		self.assertEqual(mostsec(-31536000), "365 d")
		self.assertEqual(mostsec(-31539599), "365 d")
		self.assertEqual(mostsec(-31539600), "365 d 1 h")

	def test_mostsec_str_int_seconds(self):

		self.assertEqual(mostsec("0", positive=True), None)
		self.assertEqual(mostsec("0"), "<1 ns")
		self.assertEqual(mostsec("1"), "1 s")
		self.assertEqual(mostsec("59"), "59 s")
		self.assertEqual(mostsec("60"), "1 m")
		self.assertEqual(mostsec("61"), "1 m 1 s")
		self.assertEqual(mostsec("119"), "1 m 59 s")
		self.assertEqual(mostsec("120"), "2 m")
		self.assertEqual(mostsec("120"), "2 m")
		self.assertEqual(mostsec("3599"), "59 m 59 s")
		self.assertEqual(mostsec("3600"), "1 h")
		self.assertEqual(mostsec("3601"), "1 h")
		self.assertEqual(mostsec("3659"), "1 h")
		self.assertEqual(mostsec("3660"), "1 h 1 m")
		self.assertEqual(mostsec("86399"), "23 h 59 m")
		self.assertEqual(mostsec("86400"), "1 d")
		self.assertEqual(mostsec("86401"), "1 d")
		self.assertEqual(mostsec("86459"), "1 d")
		self.assertEqual(mostsec("89999"), "1 d")
		self.assertEqual(mostsec("90000"), "1 d 1 h")
		self.assertEqual(mostsec("90000"), "1 d 1 h")
		self.assertEqual(mostsec("31535999"), "364 d 23 h")
		self.assertEqual(mostsec("31536000"), "365 d")
		self.assertEqual(mostsec("31539599"), "365 d")
		self.assertEqual(mostsec("31539600"), "365 d 1 h")

	def test_mostsec_str_int_seconds_negative(self):

		self.assertEqual(mostsec("-0", positive=True), None)
		self.assertEqual(mostsec("-0"), "<1 ns")
		self.assertEqual(mostsec("-1"), "1 s")
		self.assertEqual(mostsec("-59"), "59 s")
		self.assertEqual(mostsec("-60"), "1 m")
		self.assertEqual(mostsec("-61"), "1 m 1 s")
		self.assertEqual(mostsec("-119"), "1 m 59 s")
		self.assertEqual(mostsec("-120"), "2 m")
		self.assertEqual(mostsec("-120"), "2 m")
		self.assertEqual(mostsec("-3599"), "59 m 59 s")
		self.assertEqual(mostsec("-3600"), "1 h")
		self.assertEqual(mostsec("-3601"), "1 h")
		self.assertEqual(mostsec("-3659"), "1 h")
		self.assertEqual(mostsec("-3660"), "1 h 1 m")
		self.assertEqual(mostsec("-86399"), "23 h 59 m")
		self.assertEqual(mostsec("-86400"), "1 d")
		self.assertEqual(mostsec("-86401"), "1 d")
		self.assertEqual(mostsec("-86459"), "1 d")
		self.assertEqual(mostsec("-89999"), "1 d")
		self.assertEqual(mostsec("-90000"), "1 d 1 h")
		self.assertEqual(mostsec("-31535999"), "364 d 23 h")
		self.assertEqual(mostsec("-31536000"), "365 d")
		self.assertEqual(mostsec("-31539599"), "365 d")
		self.assertEqual(mostsec("-31539600"), "365 d 1 h")

	def test_mostsec_float_seconds(self):

		self.assertEqual(mostsec(0., positive=True), None)
		self.assertEqual(mostsec(0.), "<1 ns")
		self.assertEqual(mostsec(1.), "1 s")
		self.assertEqual(mostsec(59.), "59 s")
		self.assertEqual(mostsec(60.), "1 m")
		self.assertEqual(mostsec(61.), "1 m 1 s")
		self.assertEqual(mostsec(119.), "1 m 59 s")
		self.assertEqual(mostsec(120.), "2 m")
		self.assertEqual(mostsec(120.), "2 m")
		self.assertEqual(mostsec(3599.), "59 m 59 s")
		self.assertEqual(mostsec(3600.), "1 h")
		self.assertEqual(mostsec(3601.), "1 h")
		self.assertEqual(mostsec(3659.), "1 h")
		self.assertEqual(mostsec(3660.), "1 h 1 m")
		self.assertEqual(mostsec(86399.), "23 h 59 m")
		self.assertEqual(mostsec(86400.), "1 d")
		self.assertEqual(mostsec(86401.), "1 d")
		self.assertEqual(mostsec(86459.), "1 d")
		self.assertEqual(mostsec(89999.), "1 d")
		self.assertEqual(mostsec(90000.), "1 d 1 h")
		self.assertEqual(mostsec(31535999.), "364 d 23 h")
		self.assertEqual(mostsec(31536000.), "365 d")
		self.assertEqual(mostsec(31539599.), "365 d")
		self.assertEqual(mostsec(31539600.), "365 d 1 h")

	def test_mostsec_float_seconds_negative(self):

		self.assertEqual(mostsec(-0., positive=True), None)
		self.assertEqual(mostsec(-0.), "<1 ns")
		self.assertEqual(mostsec(-1.), "1 s")
		self.assertEqual(mostsec(-59.), "59 s")
		self.assertEqual(mostsec(-60.), "1 m")
		self.assertEqual(mostsec(-61.), "1 m 1 s")
		self.assertEqual(mostsec(-119.), "1 m 59 s")
		self.assertEqual(mostsec(-120.), "2 m")
		self.assertEqual(mostsec(-120.), "2 m")
		self.assertEqual(mostsec(-3599.), "59 m 59 s")
		self.assertEqual(mostsec(-3600.), "1 h")
		self.assertEqual(mostsec(-3601.), "1 h")
		self.assertEqual(mostsec(-3659.), "1 h")
		self.assertEqual(mostsec(-3660.), "1 h 1 m")
		self.assertEqual(mostsec(-86399.), "23 h 59 m")
		self.assertEqual(mostsec(-86400.), "1 d")
		self.assertEqual(mostsec(-86401.), "1 d")
		self.assertEqual(mostsec(-86459.), "1 d")
		self.assertEqual(mostsec(-89999.), "1 d")
		self.assertEqual(mostsec(-90000.), "1 d 1 h")
		self.assertEqual(mostsec(-31535999.), "364 d 23 h")
		self.assertEqual(mostsec(-31536000.), "365 d")
		self.assertEqual(mostsec(-31539599.), "365 d")
		self.assertEqual(mostsec(-31539600.), "365 d 1 h")

	def test_mostsec_str_float_seconds(self):

		self.assertEqual(mostsec("0.", positive=True), None)
		self.assertEqual(mostsec("0."), "<1 ns")
		self.assertEqual(mostsec("1."), "1 s")
		self.assertEqual(mostsec("59."), "59 s")
		self.assertEqual(mostsec("60."), "1 m")
		self.assertEqual(mostsec("61."), "1 m 1 s")
		self.assertEqual(mostsec("119."), "1 m 59 s")
		self.assertEqual(mostsec("120."), "2 m")
		self.assertEqual(mostsec("120."), "2 m")
		self.assertEqual(mostsec("3599."), "59 m 59 s")
		self.assertEqual(mostsec("3600."), "1 h")
		self.assertEqual(mostsec("3601."), "1 h")
		self.assertEqual(mostsec("3659."), "1 h")
		self.assertEqual(mostsec("3660."), "1 h 1 m")
		self.assertEqual(mostsec("86399."), "23 h 59 m")
		self.assertEqual(mostsec("86400."), "1 d")
		self.assertEqual(mostsec("86401."), "1 d")
		self.assertEqual(mostsec("86459."), "1 d")
		self.assertEqual(mostsec("89999."), "1 d")
		self.assertEqual(mostsec("90000."), "1 d 1 h")
		self.assertEqual(mostsec("90000."), "1 d 1 h")
		self.assertEqual(mostsec("31535999."), "364 d 23 h")
		self.assertEqual(mostsec("31536000."), "365 d")
		self.assertEqual(mostsec("31539599."), "365 d")
		self.assertEqual(mostsec("31539600."), "365 d 1 h")

	def test_mostsec_str_float_seconds_negative(self):

		self.assertEqual(mostsec("-0.", positive=True), None)
		self.assertEqual(mostsec("-0."), "<1 ns")
		self.assertEqual(mostsec("-1."), "1 s")
		self.assertEqual(mostsec("-59."), "59 s")
		self.assertEqual(mostsec("-60."), "1 m")
		self.assertEqual(mostsec("-61."), "1 m 1 s")
		self.assertEqual(mostsec("-119."), "1 m 59 s")
		self.assertEqual(mostsec("-120."), "2 m")
		self.assertEqual(mostsec("-120."), "2 m")
		self.assertEqual(mostsec("-3599."), "59 m 59 s")
		self.assertEqual(mostsec("-3600."), "1 h")
		self.assertEqual(mostsec("-3601."), "1 h")
		self.assertEqual(mostsec("-3659."), "1 h")
		self.assertEqual(mostsec("-3660."), "1 h 1 m")
		self.assertEqual(mostsec("-86399."), "23 h 59 m")
		self.assertEqual(mostsec("-86400."), "1 d")
		self.assertEqual(mostsec("-86401."), "1 d")
		self.assertEqual(mostsec("-86459."), "1 d")
		self.assertEqual(mostsec("-89999."), "1 d")
		self.assertEqual(mostsec("-90000."), "1 d 1 h")
		self.assertEqual(mostsec("-31535999."), "364 d 23 h")
		self.assertEqual(mostsec("-31536000."), "365 d")
		self.assertEqual(mostsec("-31539599."), "365 d")
		self.assertEqual(mostsec("-31539600."), "365 d 1 h")

	def test_mostsec_float_lt_second(self):

		self.assertEqual(mostsec(.0, positive=True), None)
		self.assertEqual(mostsec(.0), "<1 ns")
		self.assertEqual(mostsec(.99999999), "999 ms")
		self.assertEqual(mostsec(.9999999), "999 ms")
		self.assertEqual(mostsec(.999999), "999 ms")
		self.assertEqual(mostsec(.99999), "999 ms")
		self.assertEqual(mostsec(.9999), "999 ms")
		self.assertEqual(mostsec(.999), "999 ms")
		self.assertEqual(mostsec(.19999999), "199 ms")
		self.assertEqual(mostsec(.1999999), "199 ms")
		self.assertEqual(mostsec(.199999), "199 ms")
		self.assertEqual(mostsec(.19999), "199 ms")
		self.assertEqual(mostsec(.1999), "199 ms")
		self.assertEqual(mostsec(.199), "199 ms")
		self.assertEqual(mostsec(.19), "190 ms")
		self.assertEqual(mostsec(.1), "100 ms")
		self.assertEqual(mostsec(.10), "100 ms")
		self.assertEqual(mostsec(.100), "100 ms")
		self.assertEqual(mostsec(.1000), "100 ms")
		self.assertEqual(mostsec(.10000), "100 ms")
		self.assertEqual(mostsec(.100000), "100 ms")
		self.assertEqual(mostsec(.1000000), "100 ms")
		self.assertEqual(mostsec(.10000000), "100 ms")
		self.assertEqual(mostsec(.09999999), "99 ms")
		self.assertEqual(mostsec(.0999999), "99 ms")
		self.assertEqual(mostsec(.099999), "99 ms")
		self.assertEqual(mostsec(.09999), "99 ms")
		self.assertEqual(mostsec(.0999), "99 ms")
		self.assertEqual(mostsec(.099), "99 ms")
		self.assertEqual(mostsec(.009), "9 ms")
		self.assertEqual(mostsec(.001), "1 ms")
		self.assertEqual(mostsec(.0010), "1 ms")
		self.assertEqual(mostsec(.00100), "1 ms")
		self.assertEqual(mostsec(.001000), "1 ms")
		self.assertEqual(mostsec(.0010000), "1 ms")
		self.assertEqual(mostsec(.00100000), "1 ms")
		self.assertEqual(mostsec(.00099999), "999 us")
		self.assertEqual(mostsec(.0009999), "999 us")
		self.assertEqual(mostsec(.000999), "999 us")
		self.assertEqual(mostsec(.00099), "990 us")
		self.assertEqual(mostsec(.0009), "900 us")
		self.assertEqual(mostsec(.0001), "100 us")
		self.assertEqual(mostsec(.00010), "100 us")
		self.assertEqual(mostsec(.000100), "100 us")
		self.assertEqual(mostsec(.0001000), "100 us")
		self.assertEqual(mostsec(.00010000), "100 us")
		self.assertEqual(mostsec(.00009999), "99 us")
		self.assertEqual(mostsec(.0000999), "99 us")
		self.assertEqual(mostsec(.000099), "99 us")
		self.assertEqual(mostsec(.000009), "9 us")
		self.assertEqual(mostsec(.000001), "1 us")
		self.assertEqual(mostsec(.0000010), "1 us")
		self.assertEqual(mostsec(.00000100), "1 us")
		self.assertEqual(mostsec(.0000009999), "999 ns")
		self.assertEqual(mostsec(.000000999), "999 ns")
		self.assertEqual(mostsec(.00000099), "990 ns")
		self.assertEqual(mostsec(.0000009), "900 ns")
		self.assertEqual(mostsec(.0000001), "100 ns")
		self.assertEqual(mostsec(.00000010), "100 ns")
		self.assertEqual(mostsec(.000000100), "100 ns")
		self.assertEqual(mostsec(.0000001000), "100 ns")
		self.assertEqual(mostsec(.00000010000), "100 ns")
		self.assertEqual(mostsec(.000000100000), "100 ns")
		self.assertEqual(mostsec(.000000099999), "99 ns")
		self.assertEqual(mostsec(.00000009999), "99 ns")
		self.assertEqual(mostsec(.0000000999), "99 ns")
		self.assertEqual(mostsec(.000000099), "99 ns")
		self.assertEqual(mostsec(.000000009), "9 ns")
		self.assertEqual(mostsec(.000000001), "1 ns")
		self.assertEqual(mostsec(.0000000001), "<1 ns")

	def test_mostsec_float_lt_second_negative(self):

		self.assertEqual(mostsec(-.0, positive=True), None)
		self.assertEqual(mostsec(-.0), "<1 ns")
		self.assertEqual(mostsec(-.99999999), "999 ms")
		self.assertEqual(mostsec(-.9999999), "999 ms")
		self.assertEqual(mostsec(-.999999), "999 ms")
		self.assertEqual(mostsec(-.99999), "999 ms")
		self.assertEqual(mostsec(-.9999), "999 ms")
		self.assertEqual(mostsec(-.999), "999 ms")
		self.assertEqual(mostsec(-.19999999), "199 ms")
		self.assertEqual(mostsec(-.1999999), "199 ms")
		self.assertEqual(mostsec(-.199999), "199 ms")
		self.assertEqual(mostsec(-.19999), "199 ms")
		self.assertEqual(mostsec(-.1999), "199 ms")
		self.assertEqual(mostsec(-.199), "199 ms")
		self.assertEqual(mostsec(-.19), "190 ms")
		self.assertEqual(mostsec(-.1), "100 ms")
		self.assertEqual(mostsec(-.10), "100 ms")
		self.assertEqual(mostsec(-.100), "100 ms")
		self.assertEqual(mostsec(-.1000), "100 ms")
		self.assertEqual(mostsec(-.10000), "100 ms")
		self.assertEqual(mostsec(-.100000), "100 ms")
		self.assertEqual(mostsec(-.1000000), "100 ms")
		self.assertEqual(mostsec(-.10000000), "100 ms")
		self.assertEqual(mostsec(-.09999999), "99 ms")
		self.assertEqual(mostsec(-.0999999), "99 ms")
		self.assertEqual(mostsec(-.099999), "99 ms")
		self.assertEqual(mostsec(-.09999), "99 ms")
		self.assertEqual(mostsec(-.0999), "99 ms")
		self.assertEqual(mostsec(-.099), "99 ms")
		self.assertEqual(mostsec(-.009), "9 ms")
		self.assertEqual(mostsec(-.001), "1 ms")
		self.assertEqual(mostsec(-.0010), "1 ms")
		self.assertEqual(mostsec(-.00100), "1 ms")
		self.assertEqual(mostsec(-.001000), "1 ms")
		self.assertEqual(mostsec(-.0010000), "1 ms")
		self.assertEqual(mostsec(-.00100000), "1 ms")
		self.assertEqual(mostsec(-.00099999), "999 us")
		self.assertEqual(mostsec(-.0009999), "999 us")
		self.assertEqual(mostsec(-.000999), "999 us")
		self.assertEqual(mostsec(-.00099), "990 us")
		self.assertEqual(mostsec(-.0009), "900 us")
		self.assertEqual(mostsec(-.0001), "100 us")
		self.assertEqual(mostsec(-.00010), "100 us")
		self.assertEqual(mostsec(-.000100), "100 us")
		self.assertEqual(mostsec(-.0001000), "100 us")
		self.assertEqual(mostsec(-.00010000), "100 us")
		self.assertEqual(mostsec(-.00009999), "99 us")
		self.assertEqual(mostsec(-.0000999), "99 us")
		self.assertEqual(mostsec(-.000099), "99 us")
		self.assertEqual(mostsec(-.000009), "9 us")
		self.assertEqual(mostsec(-.000001), "1 us")
		self.assertEqual(mostsec(-.0000010), "1 us")
		self.assertEqual(mostsec(-.00000100), "1 us")
		self.assertEqual(mostsec(-.0000009999), "999 ns")
		self.assertEqual(mostsec(-.000000999), "999 ns")
		self.assertEqual(mostsec(-.00000099), "990 ns")
		self.assertEqual(mostsec(-.0000009), "900 ns")
		self.assertEqual(mostsec(-.0000001), "100 ns")
		self.assertEqual(mostsec(-.00000010), "100 ns")
		self.assertEqual(mostsec(-.000000100), "100 ns")
		self.assertEqual(mostsec(-.0000001000), "100 ns")
		self.assertEqual(mostsec(-.00000010000), "100 ns")
		self.assertEqual(mostsec(-.000000100000), "100 ns")
		self.assertEqual(mostsec(-.000000099999), "99 ns")
		self.assertEqual(mostsec(-.00000009999), "99 ns")
		self.assertEqual(mostsec(-.0000000999), "99 ns")
		self.assertEqual(mostsec(-.000000099), "99 ns")
		self.assertEqual(mostsec(-.000000009), "9 ns")
		self.assertEqual(mostsec(-.000000001), "1 ns")
		self.assertEqual(mostsec(-.0000000001), "<1 ns")

	def test_mostsec_str_float_lt_second(self):

		self.assertEqual(mostsec(".0", positive=True), None)
		self.assertEqual(mostsec(".0"), "<1 ns")
		self.assertEqual(mostsec(".99999999"), "999 ms")
		self.assertEqual(mostsec(".9999999"), "999 ms")
		self.assertEqual(mostsec(".999999"), "999 ms")
		self.assertEqual(mostsec(".99999"), "999 ms")
		self.assertEqual(mostsec(".9999"), "999 ms")
		self.assertEqual(mostsec(".999"), "999 ms")
		self.assertEqual(mostsec(".19999999"), "199 ms")
		self.assertEqual(mostsec(".1999999"), "199 ms")
		self.assertEqual(mostsec(".199999"), "199 ms")
		self.assertEqual(mostsec(".19999"), "199 ms")
		self.assertEqual(mostsec(".1999"), "199 ms")
		self.assertEqual(mostsec(".199"), "199 ms")
		self.assertEqual(mostsec(".19"), "190 ms")
		self.assertEqual(mostsec(".1"), "100 ms")
		self.assertEqual(mostsec(".10"), "100 ms")
		self.assertEqual(mostsec(".100"), "100 ms")
		self.assertEqual(mostsec(".1000"), "100 ms")
		self.assertEqual(mostsec(".10000"), "100 ms")
		self.assertEqual(mostsec(".100000"), "100 ms")
		self.assertEqual(mostsec(".1000000"), "100 ms")
		self.assertEqual(mostsec(".10000000"), "100 ms")
		self.assertEqual(mostsec(".09999999"), "99 ms")
		self.assertEqual(mostsec(".0999999"), "99 ms")
		self.assertEqual(mostsec(".099999"), "99 ms")
		self.assertEqual(mostsec(".09999"), "99 ms")
		self.assertEqual(mostsec(".0999"), "99 ms")
		self.assertEqual(mostsec(".099"), "99 ms")
		self.assertEqual(mostsec(".009"), "9 ms")
		self.assertEqual(mostsec(".001"), "1 ms")
		self.assertEqual(mostsec(".0010"), "1 ms")
		self.assertEqual(mostsec(".00100"), "1 ms")
		self.assertEqual(mostsec(".001000"), "1 ms")
		self.assertEqual(mostsec(".0010000"), "1 ms")
		self.assertEqual(mostsec(".00100000"), "1 ms")
		self.assertEqual(mostsec(".00099999"), "999 us")
		self.assertEqual(mostsec(".0009999"), "999 us")
		self.assertEqual(mostsec(".000999"), "999 us")
		self.assertEqual(mostsec(".00099"), "990 us")
		self.assertEqual(mostsec(".0009"), "900 us")
		self.assertEqual(mostsec(".0001"), "100 us")
		self.assertEqual(mostsec(".00010"), "100 us")
		self.assertEqual(mostsec(".000100"), "100 us")
		self.assertEqual(mostsec(".0001000"), "100 us")
		self.assertEqual(mostsec(".00010000"), "100 us")
		self.assertEqual(mostsec(".00009999"), "99 us")
		self.assertEqual(mostsec(".0000999"), "99 us")
		self.assertEqual(mostsec(".000099"), "99 us")
		self.assertEqual(mostsec(".000009"), "9 us")
		self.assertEqual(mostsec(".000001"), "1 us")
		self.assertEqual(mostsec(".0000010"), "1 us")
		self.assertEqual(mostsec(".00000100"), "1 us")
		self.assertEqual(mostsec(".0000009999"), "999 ns")
		self.assertEqual(mostsec(".000000999"), "999 ns")
		self.assertEqual(mostsec(".00000099"), "990 ns")
		self.assertEqual(mostsec(".0000009"), "900 ns")
		self.assertEqual(mostsec(".0000001"), "100 ns")
		self.assertEqual(mostsec(".00000010"), "100 ns")
		self.assertEqual(mostsec(".000000100"), "100 ns")
		self.assertEqual(mostsec(".0000001000"), "100 ns")
		self.assertEqual(mostsec(".00000010000"), "100 ns")
		self.assertEqual(mostsec(".000000100000"), "100 ns")
		self.assertEqual(mostsec(".000000099999"), "99 ns")
		self.assertEqual(mostsec(".00000009999"), "99 ns")
		self.assertEqual(mostsec(".0000000999"), "99 ns")
		self.assertEqual(mostsec(".000000099"), "99 ns")
		self.assertEqual(mostsec(".000000009"), "9 ns")
		self.assertEqual(mostsec(".000000001"), "1 ns")
		self.assertEqual(mostsec(".0000000001"), "<1 ns")

	def test_mostsec_str_float_lt_second_negative(self):

		self.assertEqual(mostsec("-.0", positive=True), None)
		self.assertEqual(mostsec("-.0"), "<1 ns")
		self.assertEqual(mostsec("-.99999999"), "999 ms")
		self.assertEqual(mostsec("-.9999999"), "999 ms")
		self.assertEqual(mostsec("-.999999"), "999 ms")
		self.assertEqual(mostsec("-.99999"), "999 ms")
		self.assertEqual(mostsec("-.9999"), "999 ms")
		self.assertEqual(mostsec("-.999"), "999 ms")
		self.assertEqual(mostsec("-.19999999"), "199 ms")
		self.assertEqual(mostsec("-.1999999"), "199 ms")
		self.assertEqual(mostsec("-.199999"), "199 ms")
		self.assertEqual(mostsec("-.19999"), "199 ms")
		self.assertEqual(mostsec("-.1999"), "199 ms")
		self.assertEqual(mostsec("-.199"), "199 ms")
		self.assertEqual(mostsec("-.19"), "190 ms")
		self.assertEqual(mostsec("-.1"), "100 ms")
		self.assertEqual(mostsec("-.10"), "100 ms")
		self.assertEqual(mostsec("-.100"), "100 ms")
		self.assertEqual(mostsec("-.1000"), "100 ms")
		self.assertEqual(mostsec("-.10000"), "100 ms")
		self.assertEqual(mostsec("-.100000"), "100 ms")
		self.assertEqual(mostsec("-.1000000"), "100 ms")
		self.assertEqual(mostsec("-.10000000"), "100 ms")
		self.assertEqual(mostsec("-.09999999"), "99 ms")
		self.assertEqual(mostsec("-.0999999"), "99 ms")
		self.assertEqual(mostsec("-.099999"), "99 ms")
		self.assertEqual(mostsec("-.09999"), "99 ms")
		self.assertEqual(mostsec("-.0999"), "99 ms")
		self.assertEqual(mostsec("-.099"), "99 ms")
		self.assertEqual(mostsec("-.009"), "9 ms")
		self.assertEqual(mostsec("-.001"), "1 ms")
		self.assertEqual(mostsec("-.0010"), "1 ms")
		self.assertEqual(mostsec("-.00100"), "1 ms")
		self.assertEqual(mostsec("-.001000"), "1 ms")
		self.assertEqual(mostsec("-.0010000"), "1 ms")
		self.assertEqual(mostsec("-.00100000"), "1 ms")
		self.assertEqual(mostsec("-.00099999"), "999 us")
		self.assertEqual(mostsec("-.0009999"), "999 us")
		self.assertEqual(mostsec("-.000999"), "999 us")
		self.assertEqual(mostsec("-.00099"), "990 us")
		self.assertEqual(mostsec("-.0009"), "900 us")
		self.assertEqual(mostsec("-.0001"), "100 us")
		self.assertEqual(mostsec("-.00010"), "100 us")
		self.assertEqual(mostsec("-.000100"), "100 us")
		self.assertEqual(mostsec("-.0001000"), "100 us")
		self.assertEqual(mostsec("-.00010000"), "100 us")
		self.assertEqual(mostsec("-.00009999"), "99 us")
		self.assertEqual(mostsec("-.0000999"), "99 us")
		self.assertEqual(mostsec("-.000099"), "99 us")
		self.assertEqual(mostsec("-.000009"), "9 us")
		self.assertEqual(mostsec("-.000001"), "1 us")
		self.assertEqual(mostsec("-.0000010"), "1 us")
		self.assertEqual(mostsec("-.00000100"), "1 us")
		self.assertEqual(mostsec("-.0000009999"), "999 ns")
		self.assertEqual(mostsec("-.000000999"), "999 ns")
		self.assertEqual(mostsec("-.00000099"), "990 ns")
		self.assertEqual(mostsec("-.0000009"), "900 ns")
		self.assertEqual(mostsec("-.0000001"), "100 ns")
		self.assertEqual(mostsec("-.00000010"), "100 ns")
		self.assertEqual(mostsec("-.000000100"), "100 ns")
		self.assertEqual(mostsec("-.0000001000"), "100 ns")
		self.assertEqual(mostsec("-.00000010000"), "100 ns")
		self.assertEqual(mostsec("-.000000100000"), "100 ns")
		self.assertEqual(mostsec("-.000000099999"), "99 ns")
		self.assertEqual(mostsec("-.00000009999"), "99 ns")
		self.assertEqual(mostsec("-.0000000999"), "99 ns")
		self.assertEqual(mostsec("-.000000099"), "99 ns")
		self.assertEqual(mostsec("-.000000009"), "9 ns")
		self.assertEqual(mostsec("-.000000001"), "1 ns")
		self.assertEqual(mostsec("-.0000000001"), "<1 ns")

	def test_mostsec_float_big_seconds(self):

		self.assertEqual(mostsec(1.999), "1 s")
		self.assertEqual(mostsec(59.999), "59 s")
		self.assertEqual(mostsec(60.999), "1 m")
		self.assertEqual(mostsec(61.999), "1 m 1 s")
		self.assertEqual(mostsec(119.999), "1 m 59 s")
		self.assertEqual(mostsec(120.999), "2 m")
		self.assertEqual(mostsec(120.999), "2 m")
		self.assertEqual(mostsec(3599.999), "59 m 59 s")
		self.assertEqual(mostsec(3600.999), "1 h")
		self.assertEqual(mostsec(3601.999), "1 h")
		self.assertEqual(mostsec(3659.999), "1 h")
		self.assertEqual(mostsec(3660.999), "1 h 1 m")
		self.assertEqual(mostsec(86399.999), "23 h 59 m")
		self.assertEqual(mostsec(86400.999), "1 d")
		self.assertEqual(mostsec(86401.999), "1 d")
		self.assertEqual(mostsec(86459.999), "1 d")
		self.assertEqual(mostsec(89999.999), "1 d")
		self.assertEqual(mostsec(90000.999), "1 d 1 h")
		self.assertEqual(mostsec(31535999.999), "364 d 23 h")
		self.assertEqual(mostsec(31536000.999), "365 d")
		self.assertEqual(mostsec(31539599.999), "365 d")
		self.assertEqual(mostsec(31539600.999), "365 d 1 h")

	def test_mostsec_float_big_seconds_negative(self):

		self.assertEqual(mostsec(-1.999), "1 s")
		self.assertEqual(mostsec(-59.999), "59 s")
		self.assertEqual(mostsec(-60.999), "1 m")
		self.assertEqual(mostsec(-61.999), "1 m 1 s")
		self.assertEqual(mostsec(-119.999), "1 m 59 s")
		self.assertEqual(mostsec(-120.999), "2 m")
		self.assertEqual(mostsec(-120.999), "2 m")
		self.assertEqual(mostsec(-3599.999), "59 m 59 s")
		self.assertEqual(mostsec(-3600.999), "1 h")
		self.assertEqual(mostsec(-3601.999), "1 h")
		self.assertEqual(mostsec(-3659.999), "1 h")
		self.assertEqual(mostsec(-3660.999), "1 h 1 m")
		self.assertEqual(mostsec(-86399.999), "23 h 59 m")
		self.assertEqual(mostsec(-86400.999), "1 d")
		self.assertEqual(mostsec(-86401.999), "1 d")
		self.assertEqual(mostsec(-86459.999), "1 d")
		self.assertEqual(mostsec(-89999.999), "1 d")
		self.assertEqual(mostsec(-90000.999), "1 d 1 h")
		self.assertEqual(mostsec(-31535999.999), "364 d 23 h")
		self.assertEqual(mostsec(-31536000.999), "365 d")
		self.assertEqual(mostsec(-31539599.999), "365 d")
		self.assertEqual(mostsec(-31539600.999), "365 d 1 h")

	def test_mostsec_str_float_big_seconds(self):

		self.assertEqual(mostsec("1.999"), "1 s")
		self.assertEqual(mostsec("59.999"), "59 s")
		self.assertEqual(mostsec("60.999"), "1 m")
		self.assertEqual(mostsec("61.999"), "1 m 1 s")
		self.assertEqual(mostsec("119.999"), "1 m 59 s")
		self.assertEqual(mostsec("120.999"), "2 m")
		self.assertEqual(mostsec("120.999"), "2 m")
		self.assertEqual(mostsec("3599.999"), "59 m 59 s")
		self.assertEqual(mostsec("3600.999"), "1 h")
		self.assertEqual(mostsec("3601.999"), "1 h")
		self.assertEqual(mostsec("3659.999"), "1 h")
		self.assertEqual(mostsec("3660.999"), "1 h 1 m")
		self.assertEqual(mostsec("86399.999"), "23 h 59 m")
		self.assertEqual(mostsec("86400.999"), "1 d")
		self.assertEqual(mostsec("86401.999"), "1 d")
		self.assertEqual(mostsec("86459.999"), "1 d")
		self.assertEqual(mostsec("89999.999"), "1 d")
		self.assertEqual(mostsec("90000.999"), "1 d 1 h")
		self.assertEqual(mostsec("31535999.999"), "364 d 23 h")
		self.assertEqual(mostsec("31536000.999"), "365 d")
		self.assertEqual(mostsec("31539599.999"), "365 d")
		self.assertEqual(mostsec("31539600.999"), "365 d 1 h")

	def test_mostsec_str_float_big_seconds_negative(self):

		self.assertEqual(mostsec("-1.999"), "1 s")
		self.assertEqual(mostsec("-59.999"), "59 s")
		self.assertEqual(mostsec("-60.999"), "1 m")
		self.assertEqual(mostsec("-61.999"), "1 m 1 s")
		self.assertEqual(mostsec("-119.999"), "1 m 59 s")
		self.assertEqual(mostsec("-120.999"), "2 m")
		self.assertEqual(mostsec("-120.999"), "2 m")
		self.assertEqual(mostsec("-3599.999"), "59 m 59 s")
		self.assertEqual(mostsec("-3600.999"), "1 h")
		self.assertEqual(mostsec("-3601.999"), "1 h")
		self.assertEqual(mostsec("-3659.999"), "1 h")
		self.assertEqual(mostsec("-3660.999"), "1 h 1 m")
		self.assertEqual(mostsec("-86399.999"), "23 h 59 m")
		self.assertEqual(mostsec("-86400.999"), "1 d")
		self.assertEqual(mostsec("-86401.999"), "1 d")
		self.assertEqual(mostsec("-86459.999"), "1 d")
		self.assertEqual(mostsec("-89999.999"), "1 d")
		self.assertEqual(mostsec("-90000.999"), "1 d 1 h")
		self.assertEqual(mostsec("-31535999.999"), "364 d 23 h")
		self.assertEqual(mostsec("-31536000.999"), "365 d")
		self.assertEqual(mostsec("-31539599.999"), "365 d")
		self.assertEqual(mostsec("-31539600.999"), "365 d 1 h")

	def test_mostsec_E_seconds(self):

		self.assertEqual(mostsec(1E5), "1 d 3 h")
		self.assertEqual(mostsec(-1E5), "1 d 3 h")
		self.assertEqual(mostsec("1E5"), "1 d 3 h")
		self.assertEqual(mostsec("-1E5"), "1 d 3 h")

	def test_mostsec_invalids(self):

		self.assertIsNone(mostsec("integer"))
		self.assertIsNone(mostsec(None))
		self.assertIsNone(mostsec([ 42 ]))
		self.assertIsNone(mostsec({ 42 }))
		self.assertIsNone(mostsec({ "value": 42 }))
		self.assertIsNone(mostsec(( 42, )))
		self.assertIsNone(mostsec(Transmutable))








	def test_mostsecfmt_int_seconds(self):

		self.assertEqual(self.posfmter(0), None)
		self.assertEqual(self.secfmter(0), "<1 ns")
		self.assertEqual(self.secfmter(1), "1 s")
		self.assertEqual(self.secfmter(59), "59 s")
		self.assertEqual(self.secfmter(60), "1 m")
		self.assertEqual(self.secfmter(61), "1 m 1 s")
		self.assertEqual(self.secfmter(119), "1 m 59 s")
		self.assertEqual(self.secfmter(120), "2 m")
		self.assertEqual(self.secfmter(120), "2 m")
		self.assertEqual(self.secfmter(3599), "59 m 59 s")
		self.assertEqual(self.secfmter(3600), "1 h")
		self.assertEqual(self.secfmter(3601), "1 h")
		self.assertEqual(self.secfmter(3659), "1 h")
		self.assertEqual(self.secfmter(3660), "1 h 1 m")
		self.assertEqual(self.secfmter(86399), "23 h 59 m")
		self.assertEqual(self.secfmter(86400), "1 d")
		self.assertEqual(self.secfmter(86401), "1 d")
		self.assertEqual(self.secfmter(86459), "1 d")
		self.assertEqual(self.secfmter(89999), "1 d")
		self.assertEqual(self.secfmter(90000), "1 d 1 h")
		self.assertEqual(self.secfmter(31535999), "364 d 23 h")
		self.assertEqual(self.secfmter(31536000), "365 d")
		self.assertEqual(self.secfmter(31539599), "365 d")
		self.assertEqual(self.secfmter(31539600), "365 d 1 h")

	def test_mostsecfmt_int_seconds_negative(self):

		self.assertEqual(self.posfmter(-0), None)
		self.assertEqual(self.secfmter(-0), "<1 ns")
		self.assertEqual(self.secfmter(-1), "1 s")
		self.assertEqual(self.secfmter(-59), "59 s")
		self.assertEqual(self.secfmter(-60), "1 m")
		self.assertEqual(self.secfmter(-61), "1 m 1 s")
		self.assertEqual(self.secfmter(-119), "1 m 59 s")
		self.assertEqual(self.secfmter(-120), "2 m")
		self.assertEqual(self.secfmter(-120), "2 m")
		self.assertEqual(self.secfmter(-3599), "59 m 59 s")
		self.assertEqual(self.secfmter(-3600), "1 h")
		self.assertEqual(self.secfmter(-3601), "1 h")
		self.assertEqual(self.secfmter(-3659), "1 h")
		self.assertEqual(self.secfmter(-3660), "1 h 1 m")
		self.assertEqual(self.secfmter(-86399), "23 h 59 m")
		self.assertEqual(self.secfmter(-86400), "1 d")
		self.assertEqual(self.secfmter(-86401), "1 d")
		self.assertEqual(self.secfmter(-86459), "1 d")
		self.assertEqual(self.secfmter(-89999), "1 d")
		self.assertEqual(self.secfmter(-90000), "1 d 1 h")
		self.assertEqual(self.secfmter(-31535999), "364 d 23 h")
		self.assertEqual(self.secfmter(-31536000), "365 d")
		self.assertEqual(self.secfmter(-31539599), "365 d")
		self.assertEqual(self.secfmter(-31539600), "365 d 1 h")

	def test_mostsecfmt_str_int_seconds(self):

		self.assertEqual(self.posfmter("0"), None)
		self.assertEqual(self.secfmter("0"), "<1 ns")
		self.assertEqual(self.secfmter("1"), "1 s")
		self.assertEqual(self.secfmter("59"), "59 s")
		self.assertEqual(self.secfmter("60"), "1 m")
		self.assertEqual(self.secfmter("61"), "1 m 1 s")
		self.assertEqual(self.secfmter("119"), "1 m 59 s")
		self.assertEqual(self.secfmter("120"), "2 m")
		self.assertEqual(self.secfmter("120"), "2 m")
		self.assertEqual(self.secfmter("3599"), "59 m 59 s")
		self.assertEqual(self.secfmter("3600"), "1 h")
		self.assertEqual(self.secfmter("3601"), "1 h")
		self.assertEqual(self.secfmter("3659"), "1 h")
		self.assertEqual(self.secfmter("3660"), "1 h 1 m")
		self.assertEqual(self.secfmter("86399"), "23 h 59 m")
		self.assertEqual(self.secfmter("86400"), "1 d")
		self.assertEqual(self.secfmter("86401"), "1 d")
		self.assertEqual(self.secfmter("86459"), "1 d")
		self.assertEqual(self.secfmter("89999"), "1 d")
		self.assertEqual(self.secfmter("90000"), "1 d 1 h")
		self.assertEqual(self.secfmter("31535999"), "364 d 23 h")
		self.assertEqual(self.secfmter("31536000"), "365 d")
		self.assertEqual(self.secfmter("31539599"), "365 d")
		self.assertEqual(self.secfmter("31539600"), "365 d 1 h")

	def test_mostsecfmt_str_int_seconds_negative(self):

		self.assertEqual(self.posfmter("-0"), None)
		self.assertEqual(self.secfmter("-0"), "<1 ns")
		self.assertEqual(self.secfmter("-1"), "1 s")
		self.assertEqual(self.secfmter("-59"), "59 s")
		self.assertEqual(self.secfmter("-60"), "1 m")
		self.assertEqual(self.secfmter("-61"), "1 m 1 s")
		self.assertEqual(self.secfmter("-119"), "1 m 59 s")
		self.assertEqual(self.secfmter("-120"), "2 m")
		self.assertEqual(self.secfmter("-120"), "2 m")
		self.assertEqual(self.secfmter("-3599"), "59 m 59 s")
		self.assertEqual(self.secfmter("-3600"), "1 h")
		self.assertEqual(self.secfmter("-3601"), "1 h")
		self.assertEqual(self.secfmter("-3659"), "1 h")
		self.assertEqual(self.secfmter("-3660"), "1 h 1 m")
		self.assertEqual(self.secfmter("-86399"), "23 h 59 m")
		self.assertEqual(self.secfmter("-86400"), "1 d")
		self.assertEqual(self.secfmter("-86401"), "1 d")
		self.assertEqual(self.secfmter("-86459"), "1 d")
		self.assertEqual(self.secfmter("-89999"), "1 d")
		self.assertEqual(self.secfmter("-90000"), "1 d 1 h")
		self.assertEqual(self.secfmter("-31535999"), "364 d 23 h")
		self.assertEqual(self.secfmter("-31536000"), "365 d")
		self.assertEqual(self.secfmter("-31539599"), "365 d")
		self.assertEqual(self.secfmter("-31539600"), "365 d 1 h")

	def test_mostsecfmt_float_seconds(self):

		self.assertEqual(self.posfmter(0.), None)
		self.assertEqual(self.secfmter(0.), "<1 ns")
		self.assertEqual(self.secfmter(1.), "1 s")
		self.assertEqual(self.secfmter(59.), "59 s")
		self.assertEqual(self.secfmter(60.), "1 m")
		self.assertEqual(self.secfmter(61.), "1 m 1 s")
		self.assertEqual(self.secfmter(119.), "1 m 59 s")
		self.assertEqual(self.secfmter(120.), "2 m")
		self.assertEqual(self.secfmter(120.), "2 m")
		self.assertEqual(self.secfmter(3599.), "59 m 59 s")
		self.assertEqual(self.secfmter(3600.), "1 h")
		self.assertEqual(self.secfmter(3601.), "1 h")
		self.assertEqual(self.secfmter(3659.), "1 h")
		self.assertEqual(self.secfmter(3660.), "1 h 1 m")
		self.assertEqual(self.secfmter(86399.), "23 h 59 m")
		self.assertEqual(self.secfmter(86400.), "1 d")
		self.assertEqual(self.secfmter(86401.), "1 d")
		self.assertEqual(self.secfmter(86459.), "1 d")
		self.assertEqual(self.secfmter(89999.), "1 d")
		self.assertEqual(self.secfmter(90000.), "1 d 1 h")
		self.assertEqual(self.secfmter(31535999.), "364 d 23 h")
		self.assertEqual(self.secfmter(31536000.), "365 d")
		self.assertEqual(self.secfmter(31539599.), "365 d")
		self.assertEqual(self.secfmter(31539600.), "365 d 1 h")

	def test_mostsecfmt_float_seconds_negative(self):

		self.assertEqual(self.posfmter(-0.), None)
		self.assertEqual(self.secfmter(-0.), "<1 ns")
		self.assertEqual(self.secfmter(-1.), "1 s")
		self.assertEqual(self.secfmter(-59.), "59 s")
		self.assertEqual(self.secfmter(-60.), "1 m")
		self.assertEqual(self.secfmter(-61.), "1 m 1 s")
		self.assertEqual(self.secfmter(-119.), "1 m 59 s")
		self.assertEqual(self.secfmter(-120.), "2 m")
		self.assertEqual(self.secfmter(-120.), "2 m")
		self.assertEqual(self.secfmter(-3599.), "59 m 59 s")
		self.assertEqual(self.secfmter(-3600.), "1 h")
		self.assertEqual(self.secfmter(-3601.), "1 h")
		self.assertEqual(self.secfmter(-3659.), "1 h")
		self.assertEqual(self.secfmter(-3660.), "1 h 1 m")
		self.assertEqual(self.secfmter(-86399.), "23 h 59 m")
		self.assertEqual(self.secfmter(-86400.), "1 d")
		self.assertEqual(self.secfmter(-86401.), "1 d")
		self.assertEqual(self.secfmter(-86459.), "1 d")
		self.assertEqual(self.secfmter(-89999.), "1 d")
		self.assertEqual(self.secfmter(-90000.), "1 d 1 h")
		self.assertEqual(self.secfmter(-31535999.), "364 d 23 h")
		self.assertEqual(self.secfmter(-31536000.), "365 d")
		self.assertEqual(self.secfmter(-31539599.), "365 d")
		self.assertEqual(self.secfmter(-31539600.), "365 d 1 h")

	def test_mostsecfmt_str_float_seconds(self):

		self.assertEqual(self.posfmter("0."), None)
		self.assertEqual(self.secfmter("0."), "<1 ns")
		self.assertEqual(self.secfmter("1."), "1 s")
		self.assertEqual(self.secfmter("59."), "59 s")
		self.assertEqual(self.secfmter("60."), "1 m")
		self.assertEqual(self.secfmter("61."), "1 m 1 s")
		self.assertEqual(self.secfmter("119."), "1 m 59 s")
		self.assertEqual(self.secfmter("120."), "2 m")
		self.assertEqual(self.secfmter("120."), "2 m")
		self.assertEqual(self.secfmter("3599."), "59 m 59 s")
		self.assertEqual(self.secfmter("3600."), "1 h")
		self.assertEqual(self.secfmter("3601."), "1 h")
		self.assertEqual(self.secfmter("3659."), "1 h")
		self.assertEqual(self.secfmter("3660."), "1 h 1 m")
		self.assertEqual(self.secfmter("86399."), "23 h 59 m")
		self.assertEqual(self.secfmter("86400."), "1 d")
		self.assertEqual(self.secfmter("86401."), "1 d")
		self.assertEqual(self.secfmter("86459."), "1 d")
		self.assertEqual(self.secfmter("89999."), "1 d")
		self.assertEqual(self.secfmter("90000."), "1 d 1 h")
		self.assertEqual(self.secfmter("90000."), "1 d 1 h")
		self.assertEqual(self.secfmter("31535999."), "364 d 23 h")
		self.assertEqual(self.secfmter("31536000."), "365 d")
		self.assertEqual(self.secfmter("31539599."), "365 d")
		self.assertEqual(self.secfmter("31539600."), "365 d 1 h")

	def test_mostsecfmt_str_float_seconds_negative(self):

		self.assertEqual(self.posfmter("-0."), None)
		self.assertEqual(self.secfmter("-0."), "<1 ns")
		self.assertEqual(self.secfmter("-1."), "1 s")
		self.assertEqual(self.secfmter("-59."), "59 s")
		self.assertEqual(self.secfmter("-60."), "1 m")
		self.assertEqual(self.secfmter("-61."), "1 m 1 s")
		self.assertEqual(self.secfmter("-119."), "1 m 59 s")
		self.assertEqual(self.secfmter("-120."), "2 m")
		self.assertEqual(self.secfmter("-120."), "2 m")
		self.assertEqual(self.secfmter("-3599."), "59 m 59 s")
		self.assertEqual(self.secfmter("-3600."), "1 h")
		self.assertEqual(self.secfmter("-3601."), "1 h")
		self.assertEqual(self.secfmter("-3659."), "1 h")
		self.assertEqual(self.secfmter("-3660."), "1 h 1 m")
		self.assertEqual(self.secfmter("-86399."), "23 h 59 m")
		self.assertEqual(self.secfmter("-86400."), "1 d")
		self.assertEqual(self.secfmter("-86401."), "1 d")
		self.assertEqual(self.secfmter("-86459."), "1 d")
		self.assertEqual(self.secfmter("-89999."), "1 d")
		self.assertEqual(self.secfmter("-90000."), "1 d 1 h")
		self.assertEqual(self.secfmter("-31535999."), "364 d 23 h")
		self.assertEqual(self.secfmter("-31536000."), "365 d")
		self.assertEqual(self.secfmter("-31539599."), "365 d")
		self.assertEqual(self.secfmter("-31539600."), "365 d 1 h")

	def test_mostsecfmt_float_lt_second(self):

		self.assertEqual(self.posfmter(.0), None)
		self.assertEqual(self.secfmter(.0), "<1 ns")
		self.assertEqual(self.secfmter(.99999999), "999 ms")
		self.assertEqual(self.secfmter(.9999999), "999 ms")
		self.assertEqual(self.secfmter(.999999), "999 ms")
		self.assertEqual(self.secfmter(.99999), "999 ms")
		self.assertEqual(self.secfmter(.9999), "999 ms")
		self.assertEqual(self.secfmter(.999), "999 ms")
		self.assertEqual(self.secfmter(.19999999), "199 ms")
		self.assertEqual(self.secfmter(.1999999), "199 ms")
		self.assertEqual(self.secfmter(.199999), "199 ms")
		self.assertEqual(self.secfmter(.19999), "199 ms")
		self.assertEqual(self.secfmter(.1999), "199 ms")
		self.assertEqual(self.secfmter(.199), "199 ms")
		self.assertEqual(self.secfmter(.19), "190 ms")
		self.assertEqual(self.secfmter(.1), "100 ms")
		self.assertEqual(self.secfmter(.10), "100 ms")
		self.assertEqual(self.secfmter(.100), "100 ms")
		self.assertEqual(self.secfmter(.1000), "100 ms")
		self.assertEqual(self.secfmter(.10000), "100 ms")
		self.assertEqual(self.secfmter(.100000), "100 ms")
		self.assertEqual(self.secfmter(.1000000), "100 ms")
		self.assertEqual(self.secfmter(.10000000), "100 ms")
		self.assertEqual(self.secfmter(.09999999), "99 ms")
		self.assertEqual(self.secfmter(.0999999), "99 ms")
		self.assertEqual(self.secfmter(.099999), "99 ms")
		self.assertEqual(self.secfmter(.09999), "99 ms")
		self.assertEqual(self.secfmter(.0999), "99 ms")
		self.assertEqual(self.secfmter(.099), "99 ms")
		self.assertEqual(self.secfmter(.009), "9 ms")
		self.assertEqual(self.secfmter(.001), "1 ms")
		self.assertEqual(self.secfmter(.0010), "1 ms")
		self.assertEqual(self.secfmter(.00100), "1 ms")
		self.assertEqual(self.secfmter(.001000), "1 ms")
		self.assertEqual(self.secfmter(.0010000), "1 ms")
		self.assertEqual(self.secfmter(.00100000), "1 ms")
		self.assertEqual(self.secfmter(.00099999), "999 us")
		self.assertEqual(self.secfmter(.0009999), "999 us")
		self.assertEqual(self.secfmter(.000999), "999 us")
		self.assertEqual(self.secfmter(.00099), "990 us")
		self.assertEqual(self.secfmter(.0009), "900 us")
		self.assertEqual(self.secfmter(.0001), "100 us")
		self.assertEqual(self.secfmter(.00010), "100 us")
		self.assertEqual(self.secfmter(.000100), "100 us")
		self.assertEqual(self.secfmter(.0001000), "100 us")
		self.assertEqual(self.secfmter(.00010000), "100 us")
		self.assertEqual(self.secfmter(.00009999), "99 us")
		self.assertEqual(self.secfmter(.0000999), "99 us")
		self.assertEqual(self.secfmter(.000099), "99 us")
		self.assertEqual(self.secfmter(.000009), "9 us")
		self.assertEqual(self.secfmter(.000001), "1 us")
		self.assertEqual(self.secfmter(.0000010), "1 us")
		self.assertEqual(self.secfmter(.00000100), "1 us")
		self.assertEqual(self.secfmter(.0000009999), "999 ns")
		self.assertEqual(self.secfmter(.000000999), "999 ns")
		self.assertEqual(self.secfmter(.00000099), "990 ns")
		self.assertEqual(self.secfmter(.0000009), "900 ns")
		self.assertEqual(self.secfmter(.0000001), "100 ns")
		self.assertEqual(self.secfmter(.00000010), "100 ns")
		self.assertEqual(self.secfmter(.000000100), "100 ns")
		self.assertEqual(self.secfmter(.0000001000), "100 ns")
		self.assertEqual(self.secfmter(.00000010000), "100 ns")
		self.assertEqual(self.secfmter(.000000100000), "100 ns")
		self.assertEqual(self.secfmter(.000000099999), "99 ns")
		self.assertEqual(self.secfmter(.00000009999), "99 ns")
		self.assertEqual(self.secfmter(.0000000999), "99 ns")
		self.assertEqual(self.secfmter(.000000099), "99 ns")
		self.assertEqual(self.secfmter(.000000009), "9 ns")
		self.assertEqual(self.secfmter(.000000001), "1 ns")
		self.assertEqual(self.secfmter(.0000000001), "<1 ns")

	def test_mostsecfmt_float_lt_second_negative(self):

		self.assertEqual(self.posfmter(-.0), None)
		self.assertEqual(self.secfmter(-.0), "<1 ns")
		self.assertEqual(self.secfmter(-.99999999), "999 ms")
		self.assertEqual(self.secfmter(-.9999999), "999 ms")
		self.assertEqual(self.secfmter(-.999999), "999 ms")
		self.assertEqual(self.secfmter(-.99999), "999 ms")
		self.assertEqual(self.secfmter(-.9999), "999 ms")
		self.assertEqual(self.secfmter(-.999), "999 ms")
		self.assertEqual(self.secfmter(-.19999999), "199 ms")
		self.assertEqual(self.secfmter(-.1999999), "199 ms")
		self.assertEqual(self.secfmter(-.199999), "199 ms")
		self.assertEqual(self.secfmter(-.19999), "199 ms")
		self.assertEqual(self.secfmter(-.1999), "199 ms")
		self.assertEqual(self.secfmter(-.199), "199 ms")
		self.assertEqual(self.secfmter(-.19), "190 ms")
		self.assertEqual(self.secfmter(-.1), "100 ms")
		self.assertEqual(self.secfmter(-.10), "100 ms")
		self.assertEqual(self.secfmter(-.100), "100 ms")
		self.assertEqual(self.secfmter(-.1000), "100 ms")
		self.assertEqual(self.secfmter(-.10000), "100 ms")
		self.assertEqual(self.secfmter(-.100000), "100 ms")
		self.assertEqual(self.secfmter(-.1000000), "100 ms")
		self.assertEqual(self.secfmter(-.10000000), "100 ms")
		self.assertEqual(self.secfmter(-.09999999), "99 ms")
		self.assertEqual(self.secfmter(-.0999999), "99 ms")
		self.assertEqual(self.secfmter(-.099999), "99 ms")
		self.assertEqual(self.secfmter(-.09999), "99 ms")
		self.assertEqual(self.secfmter(-.0999), "99 ms")
		self.assertEqual(self.secfmter(-.099), "99 ms")
		self.assertEqual(self.secfmter(-.009), "9 ms")
		self.assertEqual(self.secfmter(-.001), "1 ms")
		self.assertEqual(self.secfmter(-.0010), "1 ms")
		self.assertEqual(self.secfmter(-.00100), "1 ms")
		self.assertEqual(self.secfmter(-.001000), "1 ms")
		self.assertEqual(self.secfmter(-.0010000), "1 ms")
		self.assertEqual(self.secfmter(-.00100000), "1 ms")
		self.assertEqual(self.secfmter(-.00099999), "999 us")
		self.assertEqual(self.secfmter(-.0009999), "999 us")
		self.assertEqual(self.secfmter(-.000999), "999 us")
		self.assertEqual(self.secfmter(-.00099), "990 us")
		self.assertEqual(self.secfmter(-.0009), "900 us")
		self.assertEqual(self.secfmter(-.0001), "100 us")
		self.assertEqual(self.secfmter(-.00010), "100 us")
		self.assertEqual(self.secfmter(-.000100), "100 us")
		self.assertEqual(self.secfmter(-.0001000), "100 us")
		self.assertEqual(self.secfmter(-.00010000), "100 us")
		self.assertEqual(self.secfmter(-.00009999), "99 us")
		self.assertEqual(self.secfmter(-.0000999), "99 us")
		self.assertEqual(self.secfmter(-.000099), "99 us")
		self.assertEqual(self.secfmter(-.000009), "9 us")
		self.assertEqual(self.secfmter(-.000001), "1 us")
		self.assertEqual(self.secfmter(-.0000010), "1 us")
		self.assertEqual(self.secfmter(-.00000100), "1 us")
		self.assertEqual(self.secfmter(-.0000009999), "999 ns")
		self.assertEqual(self.secfmter(-.000000999), "999 ns")
		self.assertEqual(self.secfmter(-.00000099), "990 ns")
		self.assertEqual(self.secfmter(-.0000009), "900 ns")
		self.assertEqual(self.secfmter(-.0000001), "100 ns")
		self.assertEqual(self.secfmter(-.00000010), "100 ns")
		self.assertEqual(self.secfmter(-.000000100), "100 ns")
		self.assertEqual(self.secfmter(-.0000001000), "100 ns")
		self.assertEqual(self.secfmter(-.00000010000), "100 ns")
		self.assertEqual(self.secfmter(-.000000100000), "100 ns")
		self.assertEqual(self.secfmter(-.000000099999), "99 ns")
		self.assertEqual(self.secfmter(-.00000009999), "99 ns")
		self.assertEqual(self.secfmter(-.0000000999), "99 ns")
		self.assertEqual(self.secfmter(-.000000099), "99 ns")
		self.assertEqual(self.secfmter(-.000000009), "9 ns")
		self.assertEqual(self.secfmter(-.000000001), "1 ns")
		self.assertEqual(self.secfmter(-.0000000001), "<1 ns")

	def test_mostsecfmt_str_float_lt_second(self):

		self.assertEqual(self.posfmter(".0"), None)
		self.assertEqual(self.secfmter(".0"), "<1 ns")
		self.assertEqual(self.secfmter(".99999999"), "999 ms")
		self.assertEqual(self.secfmter(".9999999"), "999 ms")
		self.assertEqual(self.secfmter(".999999"), "999 ms")
		self.assertEqual(self.secfmter(".99999"), "999 ms")
		self.assertEqual(self.secfmter(".9999"), "999 ms")
		self.assertEqual(self.secfmter(".999"), "999 ms")
		self.assertEqual(self.secfmter(".19999999"), "199 ms")
		self.assertEqual(self.secfmter(".1999999"), "199 ms")
		self.assertEqual(self.secfmter(".199999"), "199 ms")
		self.assertEqual(self.secfmter(".19999"), "199 ms")
		self.assertEqual(self.secfmter(".1999"), "199 ms")
		self.assertEqual(self.secfmter(".199"), "199 ms")
		self.assertEqual(self.secfmter(".19"), "190 ms")
		self.assertEqual(self.secfmter(".1"), "100 ms")
		self.assertEqual(self.secfmter(".10"), "100 ms")
		self.assertEqual(self.secfmter(".100"), "100 ms")
		self.assertEqual(self.secfmter(".1000"), "100 ms")
		self.assertEqual(self.secfmter(".10000"), "100 ms")
		self.assertEqual(self.secfmter(".100000"), "100 ms")
		self.assertEqual(self.secfmter(".1000000"), "100 ms")
		self.assertEqual(self.secfmter(".10000000"), "100 ms")
		self.assertEqual(self.secfmter(".09999999"), "99 ms")
		self.assertEqual(self.secfmter(".0999999"), "99 ms")
		self.assertEqual(self.secfmter(".099999"), "99 ms")
		self.assertEqual(self.secfmter(".09999"), "99 ms")
		self.assertEqual(self.secfmter(".0999"), "99 ms")
		self.assertEqual(self.secfmter(".099"), "99 ms")
		self.assertEqual(self.secfmter(".009"), "9 ms")
		self.assertEqual(self.secfmter(".001"), "1 ms")
		self.assertEqual(self.secfmter(".0010"), "1 ms")
		self.assertEqual(self.secfmter(".00100"), "1 ms")
		self.assertEqual(self.secfmter(".001000"), "1 ms")
		self.assertEqual(self.secfmter(".0010000"), "1 ms")
		self.assertEqual(self.secfmter(".00100000"), "1 ms")
		self.assertEqual(self.secfmter(".00099999"), "999 us")
		self.assertEqual(self.secfmter(".0009999"), "999 us")
		self.assertEqual(self.secfmter(".000999"), "999 us")
		self.assertEqual(self.secfmter(".00099"), "990 us")
		self.assertEqual(self.secfmter(".0009"), "900 us")
		self.assertEqual(self.secfmter(".0001"), "100 us")
		self.assertEqual(self.secfmter(".00010"), "100 us")
		self.assertEqual(self.secfmter(".000100"), "100 us")
		self.assertEqual(self.secfmter(".0001000"), "100 us")
		self.assertEqual(self.secfmter(".00010000"), "100 us")
		self.assertEqual(self.secfmter(".00009999"), "99 us")
		self.assertEqual(self.secfmter(".0000999"), "99 us")
		self.assertEqual(self.secfmter(".000099"), "99 us")
		self.assertEqual(self.secfmter(".000009"), "9 us")
		self.assertEqual(self.secfmter(".000001"), "1 us")
		self.assertEqual(self.secfmter(".0000010"), "1 us")
		self.assertEqual(self.secfmter(".00000100"), "1 us")
		self.assertEqual(self.secfmter(".0000009999"), "999 ns")
		self.assertEqual(self.secfmter(".000000999"), "999 ns")
		self.assertEqual(self.secfmter(".00000099"), "990 ns")
		self.assertEqual(self.secfmter(".0000009"), "900 ns")
		self.assertEqual(self.secfmter(".0000001"), "100 ns")
		self.assertEqual(self.secfmter(".00000010"), "100 ns")
		self.assertEqual(self.secfmter(".000000100"), "100 ns")
		self.assertEqual(self.secfmter(".0000001000"), "100 ns")
		self.assertEqual(self.secfmter(".00000010000"), "100 ns")
		self.assertEqual(self.secfmter(".000000100000"), "100 ns")
		self.assertEqual(self.secfmter(".000000099999"), "99 ns")
		self.assertEqual(self.secfmter(".00000009999"), "99 ns")
		self.assertEqual(self.secfmter(".0000000999"), "99 ns")
		self.assertEqual(self.secfmter(".000000099"), "99 ns")
		self.assertEqual(self.secfmter(".000000009"), "9 ns")
		self.assertEqual(self.secfmter(".000000001"), "1 ns")
		self.assertEqual(self.secfmter(".0000000001"), "<1 ns")

	def test_mostsecfmt_str_float_lt_second_negative(self):

		self.assertEqual(self.posfmter("-.0"), None)
		self.assertEqual(self.secfmter("-.0"), "<1 ns")
		self.assertEqual(self.secfmter("-.99999999"), "999 ms")
		self.assertEqual(self.secfmter("-.9999999"), "999 ms")
		self.assertEqual(self.secfmter("-.999999"), "999 ms")
		self.assertEqual(self.secfmter("-.99999"), "999 ms")
		self.assertEqual(self.secfmter("-.9999"), "999 ms")
		self.assertEqual(self.secfmter("-.999"), "999 ms")
		self.assertEqual(self.secfmter("-.19999999"), "199 ms")
		self.assertEqual(self.secfmter("-.1999999"), "199 ms")
		self.assertEqual(self.secfmter("-.199999"), "199 ms")
		self.assertEqual(self.secfmter("-.19999"), "199 ms")
		self.assertEqual(self.secfmter("-.1999"), "199 ms")
		self.assertEqual(self.secfmter("-.199"), "199 ms")
		self.assertEqual(self.secfmter("-.19"), "190 ms")
		self.assertEqual(self.secfmter("-.1"), "100 ms")
		self.assertEqual(self.secfmter("-.10"), "100 ms")
		self.assertEqual(self.secfmter("-.100"), "100 ms")
		self.assertEqual(self.secfmter("-.1000"), "100 ms")
		self.assertEqual(self.secfmter("-.10000"), "100 ms")
		self.assertEqual(self.secfmter("-.100000"), "100 ms")
		self.assertEqual(self.secfmter("-.1000000"), "100 ms")
		self.assertEqual(self.secfmter("-.10000000"), "100 ms")
		self.assertEqual(self.secfmter("-.09999999"), "99 ms")
		self.assertEqual(self.secfmter("-.0999999"), "99 ms")
		self.assertEqual(self.secfmter("-.099999"), "99 ms")
		self.assertEqual(self.secfmter("-.09999"), "99 ms")
		self.assertEqual(self.secfmter("-.0999"), "99 ms")
		self.assertEqual(self.secfmter("-.099"), "99 ms")
		self.assertEqual(self.secfmter("-.009"), "9 ms")
		self.assertEqual(self.secfmter("-.001"), "1 ms")
		self.assertEqual(self.secfmter("-.0010"), "1 ms")
		self.assertEqual(self.secfmter("-.00100"), "1 ms")
		self.assertEqual(self.secfmter("-.001000"), "1 ms")
		self.assertEqual(self.secfmter("-.0010000"), "1 ms")
		self.assertEqual(self.secfmter("-.00100000"), "1 ms")
		self.assertEqual(self.secfmter("-.00099999"), "999 us")
		self.assertEqual(self.secfmter("-.0009999"), "999 us")
		self.assertEqual(self.secfmter("-.000999"), "999 us")
		self.assertEqual(self.secfmter("-.00099"), "990 us")
		self.assertEqual(self.secfmter("-.0009"), "900 us")
		self.assertEqual(self.secfmter("-.0001"), "100 us")
		self.assertEqual(self.secfmter("-.00010"), "100 us")
		self.assertEqual(self.secfmter("-.000100"), "100 us")
		self.assertEqual(self.secfmter("-.0001000"), "100 us")
		self.assertEqual(self.secfmter("-.00010000"), "100 us")
		self.assertEqual(self.secfmter("-.00009999"), "99 us")
		self.assertEqual(self.secfmter("-.0000999"), "99 us")
		self.assertEqual(self.secfmter("-.000099"), "99 us")
		self.assertEqual(self.secfmter("-.000009"), "9 us")
		self.assertEqual(self.secfmter("-.000001"), "1 us")
		self.assertEqual(self.secfmter("-.0000010"), "1 us")
		self.assertEqual(self.secfmter("-.00000100"), "1 us")
		self.assertEqual(self.secfmter("-.0000009999"), "999 ns")
		self.assertEqual(self.secfmter("-.000000999"), "999 ns")
		self.assertEqual(self.secfmter("-.00000099"), "990 ns")
		self.assertEqual(self.secfmter("-.0000009"), "900 ns")
		self.assertEqual(self.secfmter("-.0000001"), "100 ns")
		self.assertEqual(self.secfmter("-.00000010"), "100 ns")
		self.assertEqual(self.secfmter("-.000000100"), "100 ns")
		self.assertEqual(self.secfmter("-.0000001000"), "100 ns")
		self.assertEqual(self.secfmter("-.00000010000"), "100 ns")
		self.assertEqual(self.secfmter("-.000000100000"), "100 ns")
		self.assertEqual(self.secfmter("-.000000099999"), "99 ns")
		self.assertEqual(self.secfmter("-.00000009999"), "99 ns")
		self.assertEqual(self.secfmter("-.0000000999"), "99 ns")
		self.assertEqual(self.secfmter("-.000000099"), "99 ns")
		self.assertEqual(self.secfmter("-.000000009"), "9 ns")
		self.assertEqual(self.secfmter("-.000000001"), "1 ns")
		self.assertEqual(self.secfmter("-.0000000001"), "<1 ns")

	def test_mostsecfmt_float_big_seconds(self):

		self.assertEqual(self.secfmter(1.999), "1 s")
		self.assertEqual(self.secfmter(59.999), "59 s")
		self.assertEqual(self.secfmter(60.999), "1 m")
		self.assertEqual(self.secfmter(61.999), "1 m 1 s")
		self.assertEqual(self.secfmter(119.999), "1 m 59 s")
		self.assertEqual(self.secfmter(120.999), "2 m")
		self.assertEqual(self.secfmter(120.999), "2 m")
		self.assertEqual(self.secfmter(3599.999), "59 m 59 s")
		self.assertEqual(self.secfmter(3600.999), "1 h")
		self.assertEqual(self.secfmter(3601.999), "1 h")
		self.assertEqual(self.secfmter(3659.999), "1 h")
		self.assertEqual(self.secfmter(3660.999), "1 h 1 m")
		self.assertEqual(self.secfmter(86399.999), "23 h 59 m")
		self.assertEqual(self.secfmter(86400.999), "1 d")
		self.assertEqual(self.secfmter(86401.999), "1 d")
		self.assertEqual(self.secfmter(86459.999), "1 d")
		self.assertEqual(self.secfmter(89999.999), "1 d")
		self.assertEqual(self.secfmter(90000.999), "1 d 1 h")
		self.assertEqual(self.secfmter(31535999.999), "364 d 23 h")
		self.assertEqual(self.secfmter(31536000.999), "365 d")
		self.assertEqual(self.secfmter(31539599.999), "365 d")
		self.assertEqual(self.secfmter(31539600.999), "365 d 1 h")

	def test_mostsecfmt_float_big_seconds_negative(self):

		self.assertEqual(self.secfmter(-1.999), "1 s")
		self.assertEqual(self.secfmter(-59.999), "59 s")
		self.assertEqual(self.secfmter(-60.999), "1 m")
		self.assertEqual(self.secfmter(-61.999), "1 m 1 s")
		self.assertEqual(self.secfmter(-119.999), "1 m 59 s")
		self.assertEqual(self.secfmter(-120.999), "2 m")
		self.assertEqual(self.secfmter(-120.999), "2 m")
		self.assertEqual(self.secfmter(-3599.999), "59 m 59 s")
		self.assertEqual(self.secfmter(-3600.999), "1 h")
		self.assertEqual(self.secfmter(-3601.999), "1 h")
		self.assertEqual(self.secfmter(-3659.999), "1 h")
		self.assertEqual(self.secfmter(-3660.999), "1 h 1 m")
		self.assertEqual(self.secfmter(-86399.999), "23 h 59 m")
		self.assertEqual(self.secfmter(-86400.999), "1 d")
		self.assertEqual(self.secfmter(-86401.999), "1 d")
		self.assertEqual(self.secfmter(-86459.999), "1 d")
		self.assertEqual(self.secfmter(-89999.999), "1 d")
		self.assertEqual(self.secfmter(-90000.999), "1 d 1 h")
		self.assertEqual(self.secfmter(-31535999.999), "364 d 23 h")
		self.assertEqual(self.secfmter(-31536000.999), "365 d")
		self.assertEqual(self.secfmter(-31539599.999), "365 d")
		self.assertEqual(self.secfmter(-31539600.999), "365 d 1 h")

	def test_mostsecfmt_str_float_big_seconds(self):

		self.assertEqual(self.secfmter("1.999"), "1 s")
		self.assertEqual(self.secfmter("59.999"), "59 s")
		self.assertEqual(self.secfmter("60.999"), "1 m")
		self.assertEqual(self.secfmter("61.999"), "1 m 1 s")
		self.assertEqual(self.secfmter("119.999"), "1 m 59 s")
		self.assertEqual(self.secfmter("120.999"), "2 m")
		self.assertEqual(self.secfmter("120.999"), "2 m")
		self.assertEqual(self.secfmter("3599.999"), "59 m 59 s")
		self.assertEqual(self.secfmter("3600.999"), "1 h")
		self.assertEqual(self.secfmter("3601.999"), "1 h")
		self.assertEqual(self.secfmter("3659.999"), "1 h")
		self.assertEqual(self.secfmter("3660.999"), "1 h 1 m")
		self.assertEqual(self.secfmter("86399.999"), "23 h 59 m")
		self.assertEqual(self.secfmter("86400.999"), "1 d")
		self.assertEqual(self.secfmter("86401.999"), "1 d")
		self.assertEqual(self.secfmter("86459.999"), "1 d")
		self.assertEqual(self.secfmter("89999.999"), "1 d")
		self.assertEqual(self.secfmter("90000.999"), "1 d 1 h")
		self.assertEqual(self.secfmter("31535999.999"), "364 d 23 h")
		self.assertEqual(self.secfmter("31536000.999"), "365 d")
		self.assertEqual(self.secfmter("31539599.999"), "365 d")
		self.assertEqual(self.secfmter("31539600.999"), "365 d 1 h")

	def test_mostsecfmt_str_float_big_seconds_negative(self):

		self.assertEqual(self.secfmter("-1.999"), "1 s")
		self.assertEqual(self.secfmter("-59.999"), "59 s")
		self.assertEqual(self.secfmter("-60.999"), "1 m")
		self.assertEqual(self.secfmter("-61.999"), "1 m 1 s")
		self.assertEqual(self.secfmter("-119.999"), "1 m 59 s")
		self.assertEqual(self.secfmter("-120.999"), "2 m")
		self.assertEqual(self.secfmter("-120.999"), "2 m")
		self.assertEqual(self.secfmter("-3599.999"), "59 m 59 s")
		self.assertEqual(self.secfmter("-3600.999"), "1 h")
		self.assertEqual(self.secfmter("-3601.999"), "1 h")
		self.assertEqual(self.secfmter("-3659.999"), "1 h")
		self.assertEqual(self.secfmter("-3660.999"), "1 h 1 m")
		self.assertEqual(self.secfmter("-86399.999"), "23 h 59 m")
		self.assertEqual(self.secfmter("-86400.999"), "1 d")
		self.assertEqual(self.secfmter("-86401.999"), "1 d")
		self.assertEqual(self.secfmter("-86459.999"), "1 d")
		self.assertEqual(self.secfmter("-89999.999"), "1 d")
		self.assertEqual(self.secfmter("-90000.999"), "1 d 1 h")
		self.assertEqual(self.secfmter("-31535999.999"), "364 d 23 h")
		self.assertEqual(self.secfmter("-31536000.999"), "365 d")
		self.assertEqual(self.secfmter("-31539599.999"), "365 d")
		self.assertEqual(self.secfmter("-31539600.999"), "365 d 1 h")

	def test_mostsecfmt_E_seconds(self):

		for exponential in ( 1E5, -1E5, "1E5", "-1E5" ):
			for formatter in ( self.secfmter, self.posfmter ):
				with self.subTest(value=exponential, positive=(formatter == self.posfmter)):
					self.assertEqual(formatter(exponential), "1 d 3 h")

	def test_mostsecfmt_invalids(self):

		for curent in ( "int",None,[ 42 ],{ 42 },{ "value": 42 },( 42, ),Transmutable,print,... ):
			for formatter in ( self.secfmter, self.posfmter ):
				with self.subTest(value=curent, positive=(formatter == self.posfmter)):
					self.assertIsNone(formatter(curent))








	def test_DIRTtimer_invalids_init(self):

		for item in ( None,[ 10 ],( 10, ),{ "T": 10 },{ 10 },Transmutable,print,... ):
			with self.subTest(current=item):
				try:

					@DIRTtimer(T=item)
					class Unmanagable(Transmutable): pass

				except Exception as E:

					self.assertIsInstance(E, ValueError)
					self.assertEqual(str(E), f"Termination timer \"{item}\" is invalid")

				try:

					@DIRTtimer(D=item)
					class Unmanagable(Transmutable): pass

				except Exception as E:

					self.assertIsInstance(E, ValueError)
					self.assertEqual(str(E), f"Delay timer \"{item}\" is invalid")

				try:

					@DIRTtimer(R=item)
					class Unmanagable(Transmutable): pass

				except Exception as E:

					self.assertIsInstance(E, ValueError)
					self.assertEqual(str(E), f"Repetition counter \"{item}\" is invalid")

				try:

					@DIRTtimer(I=item)
					class Unmanagable(Transmutable): pass

				except Exception as E:

					self.assertIsInstance(E, ValueError)
					self.assertEqual(str(E), f"Interval timer \"{item}\" is invalid")

				try:

					@DIRTtimer(spawner=item)
					class Unmanagable(Transmutable): pass

				except Exception as E:

					self.assertIsInstance(E, ValueError)
					self.assertEqual(str(E), f"Process spawner \"{item}\" is invalid")




	def test_DIRTtimer_valid_inits(self):

		for item in (

			1, 900., .42,
			"500", "5.", ".00001",
			-7575483, -11., -.19999999999,
			"-0", "-00000001.", "-000000.1",
			1E5, -2E3, "14E88", "-25E17",
			True, False
		):
			with self.subTest(current=item):

				@DIRTtimer(T=item, D=item, R=item, I=item)
				class Managable(Transmutable): pass
				self.assertIsInstance(Managable, type)


		@DIRTtimer()
		class Managable(Transmutable): pass
		self.assertIsInstance(Managable, type)




	def test_DIRTtimer_dummy_spawn(self):
		class DummyProcess(Transmutable):

			def __call__(self, *args, **kwargs): return self
			class loggy(LibraryContrib):

				handler		= self.TIMETURN_TIMERS_HANDLER
				init_name	= "DIRTtimer_dummy_spawn"

			def start(self):		self.loggy.info("Dummy \"start\"")
			def terminate(self):	self.loggy.info("Dummy \"terminate\"")

		self.test_case = DummyProcess()

		@DIRTtimer(T=.42, spawner=self.test_case)
		class ScaryWoker(Transmutable):
			def __call__(self): pass

		with self.assertLogs("DIRTtimer_dummy_spawn", 20) as case_loggy : ScaryWoker()()
		self.assertIn("INFO:DIRTtimer_dummy_spawn:Dummy \"start\"", case_loggy.output)
		self.assertIn("INFO:DIRTtimer_dummy_spawn:Dummy \"terminate\"", case_loggy.output)




	@unittest.skipIf(os.name == "nt", "cannot test termination, cause windows cannot fork")
	def test_DIRTtimer_complex(self):

		DIRT_complex = str(self.MAGICAL_ROOT /"DIRT_complex.loggy")
		self.make_loggy_file(DIRT_complex)

		@DIRTtimer(T=.42, D=.69, R=2, I=.75)
		class Monday(self.browser):
			class loggy(LibraryContrib):

				handler		= DIRT_complex
				init_name	= "DIRTtimer"
				init_level	= 10

		loggy = []
		RAM = [ 1000 ]
		self.test_case = Monday()
		self.test_case("open tab", source=RAM)
		self.assertEqual(RAM,[ 1000 ])
		self.test_case.loggy.close()

		with open(DIRT_complex) as case_loggy:
			for line in case_loggy : loggy.append(line.rstrip("\n")[16:])

		self.assertEqual(

			loggy,
			[
				"@DIRTtimer-Monday DEBUG : Delay start timer 0.69 seconds",
				"@DIRTtimer-Monday DEBUG : Interval delay 0.75 seconds",
				"@DIRTtimer-Monday DEBUG : Repetition counter 2 times",
				"@DIRTtimer-Monday DEBUG : Termination timer 0.42 seconds",
				"@DIRTtimer-Monday DEBUG : Caller arguments: ('open tab',)",
				"@DIRTtimer-Monday DEBUG : Caller keyword arguments: {'source': [1000]}",
				"@DIRTtimer-Monday DEBUG : DIRT iteration 1",
				f"@DIRTtimer-Monday DEBUG : Delay 0.69 seconds for {self.test_case} performed",
				f"@DIRTtimer-Monday INFO : Starting 0.42 seconds timer for {self.test_case}",
				"@DIRTtimer-Monday INFO : Going to open tab",
				"@DIRTtimer-Monday INFO : Current RAM consumed 2000",
				f"@DIRTtimer-Monday DEBUG : Process {self.test_case} terminated",
				f"@DIRTtimer-Monday DEBUG : Interval 0.75 seconds for {self.test_case} performed",
				"@DIRTtimer-Monday DEBUG : DIRT iteration 2",
				f"@DIRTtimer-Monday DEBUG : Delay 0.69 seconds for {self.test_case} performed",
				f"@DIRTtimer-Monday INFO : Starting 0.42 seconds timer for {self.test_case}",
				"@DIRTtimer-Monday INFO : Going to open tab",
				"@DIRTtimer-Monday INFO : Current RAM consumed 2000",
				f"@DIRTtimer-Monday DEBUG : Process {self.test_case} terminated",
				f"@DIRTtimer-Monday DEBUG : Interval 0.75 seconds for {self.test_case} performed",
			]
		)
		if	os.path.isfile(DIRT_complex): os.remove(DIRT_complex)








	def test_DIRTtimer_defaults(self):

		DIRT_defaults = str(self.MAGICAL_ROOT /"DIRT_defaults.loggy")
		self.make_loggy_file(DIRT_defaults)

		@DIRTtimer()
		class Tuesday(self.browser):
			class loggy(LibraryContrib):

				handler		= DIRT_defaults
				init_name	= "DIRTtimer-d"

		loggy = []
		RAM = [ 1000 ]
		self.test_case = Tuesday()
		self.test_case("open tab", source=RAM)
		self.assertEqual(RAM,[ 2000 ])
		self.test_case.loggy.close()

		with open(DIRT_defaults) as case_loggy:
			for line in case_loggy : loggy.append(line.rstrip("\n")[16:])

		self.assertEqual(

			loggy,
			[
				"@DIRTtimer-d INFO : Going to open tab",
				"@DIRTtimer-d INFO : Current RAM consumed 2000",
			]
		)
		if	os.path.isfile(DIRT_defaults): os.remove(DIRT_defaults)








	def test_DIRTtimer_soloproc(self):

		DIRT_soloproc = str(self.MAGICAL_ROOT /"DIRT_soloproc.loggy")
		self.make_loggy_file(DIRT_soloproc)

		@DIRTtimer(R=10, I=.05)
		class Wednesday(self.browser):
			class loggy(LibraryContrib):

				handler			= DIRT_soloproc
				init_name		= "DIRTtimer-s"
				force_handover	= True

		loggy = []
		RAM = [ 1000 ]
		self.test_case = Wednesday()
		self.test_case("open tab", source=RAM)
		self.assertEqual(RAM,[ 1024000 ])
		self.test_case.loggy.close()

		with open(DIRT_soloproc) as case_loggy:
			for line in case_loggy : loggy.append(line.rstrip("\n")[16:])

		self.assertEqual(

			loggy,
			[
				"@DIRTtimer-s-Wednesday INFO : Going to open tab",
				"@DIRTtimer-s-Wednesday INFO : Current RAM consumed 2000",
				"@DIRTtimer-s-Wednesday INFO : Going to open tab",
				"@DIRTtimer-s-Wednesday INFO : Current RAM consumed 4000",
				"@DIRTtimer-s-Wednesday INFO : Going to open tab",
				"@DIRTtimer-s-Wednesday INFO : Current RAM consumed 8000",
				"@DIRTtimer-s-Wednesday INFO : Going to open tab",
				"@DIRTtimer-s-Wednesday INFO : Current RAM consumed 16000",
				"@DIRTtimer-s-Wednesday INFO : Going to open tab",
				"@DIRTtimer-s-Wednesday INFO : Current RAM consumed 32000",
				"@DIRTtimer-s-Wednesday INFO : Going to open tab",
				"@DIRTtimer-s-Wednesday INFO : Current RAM consumed 64000",
				"@DIRTtimer-s-Wednesday INFO : Going to open tab",
				"@DIRTtimer-s-Wednesday INFO : Current RAM consumed 128000",
				"@DIRTtimer-s-Wednesday INFO : Going to open tab",
				"@DIRTtimer-s-Wednesday INFO : Current RAM consumed 256000",
				"@DIRTtimer-s-Wednesday INFO : Going to open tab",
				"@DIRTtimer-s-Wednesday INFO : Current RAM consumed 512000",
				"@DIRTtimer-s-Wednesday INFO : Going to open tab",
				"@DIRTtimer-s-Wednesday INFO : Current RAM consumed 1024000",
			]
		)
		if	os.path.isfile(DIRT_soloproc): os.remove(DIRT_soloproc)








	def test_DIRTtimer_nocall(self):

		@DIRTtimer(T=1000)
		class Unmanagable(Transmutable): pass
		self.assertRaisesRegex(

			TypeError,
			"Unmanagable object is not callable",
			Unmanagable()
		)

	def test_DIRTtimer_positional_arguments(self):

		try:

			@DIRTtimer(1000,I=2000,R=3000,T=4000)
			class Unmanagable(Transmutable): pass

		except Exception as E:

			self.assertIsInstance(E, TypeError)
			self.assertEqual(

				str(E),
				"DIRTtimer.__init__() takes 1 positional argument but 2 positional arguments "
				"(and 3 keyword-only arguments) were given"
			)

		try:

			@DIRTtimer(1000,2000,R=3000,T=4000)
			class Unmanagable(Transmutable): pass

		except Exception as E:

			self.assertIsInstance(E, TypeError)
			self.assertEqual(

				str(E),
				"DIRTtimer.__init__() takes 1 positional argument but 3 positional arguments "
				"(and 2 keyword-only arguments) were given"
			)

		try:

			@DIRTtimer(1000,2000,3000,T=4000)
			class Unmanagable(Transmutable): pass

		except Exception as E:

			self.assertIsInstance(E, TypeError)
			self.assertEqual(

				str(E),
				"DIRTtimer.__init__() takes 1 positional argument but 4 positional arguments "
				"(and 1 keyword-only argument) were given"
			)

		try:

			@DIRTtimer(1000,2000,3000,4000)
			class Unmanagable(Transmutable): pass

		except Exception as E:

			self.assertIsInstance(E, TypeError)
			self.assertEqual(

				str(E),
				"DIRTtimer.__init__() takes 1 positional argument but 5 were given"
			)








if	__name__ == "__main__" : unittest.main(verbosity=2)










from typing												import Any
from typing												import List
from typing												import Hashable
from typing												import Optional
from typing												import Callable
from collections.abc									import Sequence
from pygwarts.magical.philosophers_stone				import Transmutable
from pygwarts.magical.philosophers_stone.transmutations	import ControlledTransmutation
from pygwarts.magical.spells							import geminio
from pygwarts.irma.shelve								import LibraryShelf








def shelf_case(	target	:Any,
				*,
				key		:Hashable,
				shelf	:LibraryShelf | dict,
				prep	:Callable[[Any],Any],
				post	:Callable[[Any],Any],
			)->	Any		:

	"""
		irma.shelve.casing core function for "casing", or in other words for maintaining shelves on the run.
		The idea behind this function is accepting some value "target", which is to be preprocessed some
		way, shelved and return after available postprocess. The keyword-only arguments "key" and "shelf"
		are corresponding mapping name and LibraryShelf object or dict to store "target". Before storing,
		the previous value in the "shelf" will be retrieved and preprocessed too. Assertion will be made
		for preprocessed "target" is both not None and successfully shelved. Both preprocessed "target"
		value and previous value will be passed to postprocess function and immediately returned. Any
		Exception raised on the way of returning postprocessed result will be caught and skipped, so
		None will be returned. Two last keyword-only arguments "prep" and "post" are responsible for
		preprocessing both "target" and previous values separately and postprocessing both "target" and
		previous values simultaneously. "prep" must be callable which accepts single argument, handles
		it proper way, and returns desired value to replace original value. "post" must be callable which
		accepts two arguments, current and previous values, and handles it proper way to return single
		value as a final result. "shelf" must be either LibraryShelf or a dictionary with "get" method.
	"""

	try:

		T = prep(target)
		assert T is not None


		match shelf:

			case LibraryShelf():	P = prep(shelf[key])
			case dict():			P = prep(shelf.get(key))
			case _:					return


		shelf[key] = T
		assert shelf[key] == T
		return post(T,P)


	except: return








def is_num(candidate :int|float|str) -> int | float | str | None :

	"""
		Helper function, that might be used as "prep" callable for handling numerical values. Accepts
		single argument "candidate" and performs double cast:
			1 - cast to string to exclude boolean values;
			2 - cast to float to encompass numerical values.
		If no Exception raised during double cast, it must ensure "candidate" is a numerical value, so it
		will be returned. Otherwise returns None.
	"""

	try:	float(str(candidate))
	except:	return
	else:	return candidate




def is_iterable(candidate :Sequence) -> Sequence | None :

	"""
		Helper function, that might be used as "prep" callable for handling iterable values. Accepts
		single argument "candidate" and pass it to built-in "iter" function. If not Exception raised,
		it must ensure "candidate" is iterable object, so it will be returned. Otherwise returns None.
	"""

	try:	iter(candidate)
	except:	return
	else:	return candidate








def num_diff(num1 :int|float|str, num2 :int|float|str) -> str :

	"""
		Helper function, that might be used as "post" callable for handling numerical values. Accepts
		two arguments "num1" and "num2", which must be valid numerical values. This function will put
		both arguments in built-in "eval" function in order to obtain third numerical value, as difference
		between "num1" and "num2". This difference will be converted to special string, depending on it's
		sign:
			0 <difference: (+difference);
			difference <0: (-difference).
		For difference == 0 special string will be empty. Returns final string that will looks like:
			num1[ (+-difference)]
		Doesn't handle any Exception.
	"""

	diff = eval(f"{num1}-{num2 if num2 is not None else 0}")
	diff = f"{' (' if str(diff).startswith('-') else ' (+'}{diff})" if diff else ""

	return f"{num1}{diff}"




def seq_diff(seq1 :Sequence, seq2 :Sequence) -> List[str] :

	"""
		Helper function, that might be used as "post" callable for handling iterable objects. Accepts
		two arguments "seq1" and "seq2", which must be valid iterable objects. This function will
		compare both objects and obtain a set of "seq1" objects that are not present in "seq2".
		As a result will construct a list from all "seq1" elements, converted to strings. For every
		"seq1" element that not present in "seq2", it's string will be marked with " (+)" to identify
		the "difference" elements. Returns final list of strings. It must be noticed, that despite this
		function may work for any type of items in "seq1" and "seq2", their comparison will be direct,
		but final list will always content string representations of "seq1", so basically this function
		implies return of modified "seq1".
	"""

	delta = set(seq1) - set(seq2 or [])
	return [ f"{item}{' (+)' if item in delta else ''}" for item in seq1 ]








class ShelfCase(ControlledTransmutation):

	"""
		irma.shelve utility decorator, that integrates the core "casing" functionality into mutable chain.
		As ControlledTransmutation class, accepts positional argument "link", which must be a string that
		represents the name of LibraryShelf attribute, and optional keyword only arguments "prep" and
		"post", which must be "casing" handling callables. The key for LibraryShelf will be a string
		representation of current mutable chain point, and it will be passed along with this three
		accepted arguments to "shelf_case" function. In mutable chain acts as a mutation - takes
		decorated Transmutable and overwrites it's __call__, which must return single value to be passed
		into "shelf_case" function. If "shelf_case" function returns None, that might mean it failed to
		produce final value, original value, obtained from decorated __call__ will be returned. Otherwise
		returns any value obtained from "shelf_case" function. Returns None if decorated __call__ returns
		None. This is the general type of "casing" decorator, which might be supplied with different
		"prep" and "post" for certain aims. If "shelf_case" returns None, the state of mapping in
		LibraryShelf object is undefined. By default "prep" and "post" are callables which simply returns
		it's first argument.
	"""

	def __init__(
					self,
					link	:str,
					*,
					prep	:Optional[Callable[[Any],Any]]	=(lambda T	: T),
					post	:Optional[Callable[[Any],Any]]	=(lambda T,P: T),
				):

		self.link = link
		self.prep = prep
		self.post = post

	def _mutable_chain_injection(self, layer :Transmutable) -> Transmutable :

		link = self.link
		post = self.post
		prep = self.prep


		class ShelfCasing(geminio(layer)):
			def __call__(self, *args, **kwargs) -> Any :

				if	(target := super().__call__(*args, **kwargs)) is not None:
					if	(casing := shelf_case(

						target,
						key=str(self),
						shelf=getattr(self, str(link), None),
						prep=prep,
						post=post,

					))	is not None:

						self.loggy.debug(f"Obtained shelf casing value \"{casing}\"")
						return casing


					self.loggy.debug(f"Failed to obtain shelf casing value")
					return target


		return	ShelfCasing








class NumDiffCase(ControlledTransmutation):

	"""
		irma.shelve utility decorator, that integrates special "casing" functionality into mutable chain,
		along with obtaining of special string of difference between two numerical values. As
		ControlledTransmutation class, accepts argument "link", which must be a string that represents
		the name of LibraryShelf attribute. The key for LibraryShelf will be a string representation of
		current mutable chain point, and it will be passed along with found "shelf", "is_num" prep and
		"num_diff" post to "shelf_case" function. In mutable chain acts as a mutation - takes decorated
		Transmutable and overwrites it's __call__, which must return a numerical value to be passed into
		"shelf_case" function. If "shelf_case" function returns None, that might mean it failed to validate
		either __call__ or retrieved from LibraryShelf values, or to produce final value with "post", so
		original value, obtained from decorated __call__ will be returned. Otherwise returns a string,
		corresponding to "num_diff" function. Returns None if decorated __call__ returns None.
	"""

	def __init__(self, link :str): self.link = link
	def _mutable_chain_injection(self, layer :Transmutable) -> Transmutable :

		link = self.link


		class NumDiffCasing(geminio(layer)):
			def __call__(self, *args, **kwargs) -> str | None :

				if	(target := super().__call__(*args, **kwargs)) is not None:
					if	isinstance(

						casing := shelf_case(

							target,
							key=str(self),
							shelf=getattr(self, str(link), None),
							prep=is_num,
							post=num_diff,
						),	str
					):
						self.loggy.debug(f"Obtained number difference string \"{casing}\"")
						return casing


					self.loggy.debug("Number difference not obtained")
					return target


		return	NumDiffCasing








class SeqDiffCase(ControlledTransmutation):

	"""
		irma.shelve utility decorator, that integrates special "casing" functionality into mutable chain,
		along with comparison of iterable objects. As ControlledTransmutation class, accepts argument
		"link", which must be a string that represents the name of LibraryShelf attribute. The key for
		LibraryShelf will be a string representation of current mutable chain point, and it will be passed
		along with found "shelf", "is_iterable" prep and "seq_diff" post to "shelf_case" function. In
		mutable chain acts as a mutation - takes decorated Transmutable and overwrites it's __call__, which
		must return an iterable to be passed into "shelf_case" function. If "shelf_case" function returns
		None, that might mean it failed to validate either __call__ or retrieved from LibraryShelf values,
		or to produce final value with "post", so original value, obtained from decorated __call__ will be
		returned. Otherwise returns a string, corresponding to "num_diff" function. Returns None if
		decorated __call__ returns None.
	"""

	def __init__(self, link :str): self.link = link
	def _mutable_chain_injection(self, layer :Transmutable) -> Transmutable :

		link = self.link


		class SeqDiffCasing(geminio(layer)):
			def __call__(self, *args, **kwargs) -> str | None :

				if	(target := super().__call__(*args, **kwargs)) is not None:
					if	isinstance(

						casing := shelf_case(

							target,
							key=str(self),
							shelf=getattr(self, str(link), None),
							prep=is_iterable,
							post=seq_diff,
						),	list
					):
						self.loggy.debug(f"Obtained sequence difference \"{casing}\"")
						return casing


					self.loggy.debug("Sequence difference not obtained")
					return target


		return	SeqDiffCasing










import	unittest
from	pygwarts.tests.filch.unit_filch_apppuf			import *
from	pygwarts.tests.filch.unit_filch_transportclaw	import *
from	pygwarts.tests.filch.unit_filch_nettherin		import *
from	pygwarts.tests.filch.unit_filch_linkindor		import *
from	pygwarts.tests.filch.unit_filch_marauders_map	import *








if	__name__ == "__main__" : unittest.main(verbosity=2)










from typing									import Any
from typing									import Callable
from pygwarts.magical.philosophers_stone	import Transmutable
from pygwarts.magical.spells				import patronus
from pygwarts.filch.nettherin				import is_valid_ip4
from pygwarts.filch.transportclaw			import is_valid_port








class SNMPtrap(Transmutable):

	"""
		Application layer class for serving SNMP notifications.
		Must accept setup positional arguments:
			listen_ip	- ip4 address to listen for notifications, as string;
			listen_port	- port number to listen for notifications, as int;
			trap		- callable that must accept "listen_ip" and "listen_port", along with any
						additional flags as **kwargs, and establish the actual listener.
		Once "listen_ip" and "listen_port" validated, the "trap" invocation starts. It must be
		maintained by the "trap" all the logic for SNMP processing and also the working time.
		When "trap" will be called all keyword arguments from current object __call__ will be
		passed to it, so this is the way for regulating "trap" from outside. Any Exception during
		"trap" working time will be caught and logged, and this will cause "trap" to stop.
	"""

	def __call__(
					self,
					listen_ip	:str,
					listen_port	:int,
					trap		:Callable[[Any],None],
					**kwargs
				):

		if	is_valid_ip4(listen_ip):
			if	is_valid_port(listen_port):


				self.listen_ip = listen_ip
				self.listen_port = listen_port


				self.loggy.info(f"Establishing SNMP trap at {self.listen_ip}:{self.listen_port}")
				for k,v in kwargs.items() : self.loggy.debug(f"Using {k}: {v}")


				try:	trap(self.listen_ip, self.listen_port, **kwargs)
				except	Exception as E: self.loggy.error(f"{self} got {patronus(E)}")
				else:	self.loggy.info(f"{self.listen_ip}:{self.listen_port} SNMP trap discarded")


			else:	self.loggy.debug("Listen port verification failed")
		else:		self.loggy.debug("Listen ip verification failed")










from typing									import Any
from typing									import Callable
from pygwarts.magical.philosophers_stone	import Transmutable
from pygwarts.magical.spells				import patronus
from pygwarts.filch.nettherin				import is_valid_ip4
from pygwarts.filch.linkindor				import EUI48_format








class HostDiscovery(Transmutable):

	"""
		Link layer class for discovering network hosts, by sending ARP request and processing corresponding
		response. Must accept positional arguments:
			target		- IP4 address of target host to be discovered, as string;
			discoverer	- callable that must accept "target", along with any additional flags as **kwargs,
						and implement actual host discovery whole process.
		Once "target" validated and "discoverer" assured callable, "discoverer" will be invoked with
		"target" and all keyword arguments from current object __call__. If "discoverer" call results
		valid MAC address, which must be a discovered host physical address as response, it will be
		returned. In any other cases, including Exception raise during "discoverer" call, corresponding
		logging will occur for identification and None will be returned.
	"""

	def __call__(self, target :str, discoverer :Callable[[Any],str], **kwargs) -> str | None :


		if	is_valid_ip4(target):
			if	callable(discoverer):


				self.loggy.debug(f"Discovering {target}")
				for k,v in kwargs.items() : self.loggy.debug(f"Using {k}: {v}")


				try:

					if	isinstance(response_MAC := EUI48_format(discoverer(target, **kwargs)), str):

						self.loggy.info(f"Received response for {target} at {response_MAC}")
						return response_MAC
					else:
						self.loggy.debug(f"No response from {target}")


				except	Exception as E:

					self.loggy.debug(f"Discovery failed due to {patronus(E)}")
			else:	self.loggy.debug(f"Invalid discoverer \"{discoverer}\"")
		else:		self.loggy.debug(f"Invalid IP4 address \"{target}\"")










from	__future__	import annotations
from	typing		import Optional
from	typing		import Tuple
from	typing		import List
from	typing		import Dict
from	logging		import getLogger
from	logging		import Formatter
from	logging		import StreamHandler
from	logging		import FileHandler
from	sys			import stdout
from	io			import TextIOWrapper
from	os			import path		as ospath
from	os			import makedirs
import	re








class LibraryContrib:

	"""
		                                   ..-=+++=..   
		                                ..=++-:...-#=.  
		                              .:++-.........#-. 
		                            .-*+:...:::::...+*. 
		                          .:*=....::::::::..=#. 
		                         .++:.::::::::::::..+*. 
		                       .:*-..::::::::::::...#-. 
		                      .=*..::::::=-::::::..-%.  
		                   .:=*+..:-:::=#=:::::::..#-.  
		                 .=*=++..:=::-*%=:::--::..+*=++.
		               .:#=.:+..:=-:=#%=::-=-::..=%*-.%:
		              .=#:.:=:.:-=:=##=::==::...-*:..:%.
		             .+#:.:::.::+--##+:=+-:..:.-=....+*.
		            .+#:.::::.::*:#*#=+-:::::..:....:%:.
		           .=%:.:::::.:-*+#%#=::::::....:...*+. 
		          .:%-.::::::::=###+::::::::..:::..=#.  
		          .**.:::::::::-%*#::::--:...:::..-#:   
		          :%-.::::.-:::=##=::-=-:..::::..-#--.  
		     ..   =#.::::::=:::#*#::==::.:::::..=%*##.  
		     .=+: *+.::::::=::=##=-+-::::::::..+#=.#-.  
		      -#++%=::::::.+::**%++::::::::..:#=..=#.   
		      :%:=%-::::::.*:-%*%=::::::::..=*:..:%:    
		      .%-:#=:::::::==+##+::::::::.:*+:...*+.    
		      .%=:**::::::::##+%-::::--:.=*-:...+#.     
		      .%=:-#:::::=::=%+#::::--::++::...=#.      
		      .#+::+=::::-+--#**::-=:::*=::...+#.       
		      .#+:::+::.:::+##*+:-=:::*-:::..**.        
		      .**::::-::::::**#+-+:::=-:::.:#=.         
		....   =#:::::::.:::*+#+*::::-:::.-#-.          
		.-**=-..%-:::::::..:#+#%-:::::::.=#..           
		  .=*=++##::::::::..#+#*::::::::+*:===*+:       
		   ..++:=#*::-::::::#+#=:::::::+%+=:-*-. .      
		   ...-#-:+*-:=:::::#+#+::-:::+*:::+*-=++-.     
		.-+##+++*-:-*+:==-::*+*+:=-::-#:::=*+-++:       
		   .:=*=:::::-+=-=++#***-+:::*-:::::-#-.        
		      .-*=::::::::::+*+#*-:::+:::::=#..         
		        .=*=::::::::=#=%#:::::::::=*..          
		          .-++=-::::-%=#+:::::::=*=.            
		        ..-=++*+=-:::#+*+::::=+*#==..           
		      ..=+====*%*=:::**+#::----+**++-.          
		           .-*=::::::=#=%-::::-==*#=.           
		          .-#++=====+=%=*+-++=-::....           
		          .--..    ..-#*=%*..                   
		                      :#=#-                     
		                      .#++*.                    
		                       -#=%:                    
		                       .#=+*.                   
		                        -#=#-                   
		                        .#=+#.                  
		                         -#=*=.                 
		                         .*+=#:.                
		                         ..#++#..               
		                           .:-=:.               

		This class represents the wrapper over "logging" standard library functionality. It uses minimal
		"Logger" set up either for stdout or file handling, but covers it with additional managing layer:
			modularity		- every object that accessing LibraryContrib is treated as a module. It is
							achieved by maintaining every object string representation ad identifier for
							mappings like "handover_map" that maps them with corresponding logging levels,
							or "forcing_map" that maps regular expressions to match them to activate certain
							logging level. Additional "handover_order" list storing the order each module
							made first access to LibraryContrib object.
			runtime formats	- the modularity is bound to be reflected in logging records. LibraryContrib
							field "contribfmt" is a "logging" compatible format, that must include an option
							to inject current "module" name. That means every object that want to make
							logging record must provide it's name string to be included in that logging
							record, and it is all happens in runtime.
			handover		- functionality that allows the concept of "modularity". It is mainly consist of
							eponymous method "handover" which must be called by every "module" right
							before access to the actual logging functionality. In fact, all LibraryContrib
							fields that has "handover" in names are part of handover functionality. The
							"handover_map" is being maintained right in "handover" method.
		Operates with following variables, that might be provided either as arguments or as class fields,
		or might be even omitted and defaulted in initiation:
			handler			- string that must represent file path for "FileHandler" initiating. Might be
							an absolute or relative path. Destination folder for such handler will always
							be created if does not exist, without handling e.g. PermissionError. If omitted,
							"Logger" handler is defaulted to "stdout";
			init_name		- string that must represent desired name for "Logger". This name is what usually
							"Logger" puts in log file, and what makes the difference between Loggers, so it
							is very important for some objects to not use same name Logger simultaneously,
							cause it will result a log file mess like duplication of records. If omitted,
							the default value will be some notable strings value;
			init_level		- integer value that will represent the initial Logger level. LibraryContrib
							doesn't maintain any special levels other than "logging" offers, so it's
							levels are the same:
								10 - DEBUG,
								20 - INFO,
								30 - WARNING,
								40 - ERRORS,
								50 - CRITICAL.
							This is very concise and sufficient functionality, plus, in terms of flexibility,
							LibraryContrib designed to recognize any values in between levels ranges. Default
							level is 20 - INFO;
			force_handover	- boolean value that represents a flag for LibraryContrib "handover". By default
							it's set to False and only debug level does activate handovered objects strings
							representation in records, but if "force_handover" set to True it will encompass
							every logging level, until set back to False;
			force_debug		- iterable that must content strings, which when compiled by "re" module, will
							match the "modules" names that are desired to be forced to a DEBUG logging level.
							Empty by default and if provided must not be a string;
			force_info		- iterable that must content strings, which when compiled by "re" module, will
							match the "modules" names that are desired to be forced to a INFO logging level.
							Empty by default and if provided must not be a string;
			force_warning	- iterable that must content strings, which when compiled by "re" module, will
							match the "modules" names that are desired to be forced to a WARNING logging
							level. Empty by default and if provided must not be a string;
			force_error		- iterable that must content strings, which when compiled by "re" module, will
							match the "modules" names that are desired to be forced to a ERROR logging level.
							Empty by default and if provided must not be a string;
			force_critical	- iterable that must content strings, which when compiled by "re" module, will
							match the "modules" names that are desired to be forced to a CRITICAL logging
							level. Empty by default and if provided must not be a string;
			contribfmt		- dictionary that must content "fmt" and "datefmt" strings mappings to comply
							"Formatter" patterns. The "fmt" string must always include "origin" format
							argument, that will be used for handover mentioning. The default format will
							looks like this: DD/MM/YYYY HHMM @logger-handover LEVEL : MESSAGE
		LibraryContrib might be used as a regular logger for any object, but it's real purpose is to
		work with Transmutable objects. Ability to "handover" Transmutables will always allow to track
		current object name, as suggested by Transmutable __str__. The very special field "CONTRIB_HOOK"
		by default set to True in order to allow Transmutable to hook LibraryContrib type object.
	"""


	handler			:Optional[str]
	init_name		:Optional[str]
	init_level		:Optional[int]
	force_handover	:Optional[bool]
	force_debug		:Optional[List[str] | Tuple[str]]
	force_info		:Optional[List[str] | Tuple[str]]
	force_warning	:Optional[List[str] | Tuple[str]]
	force_error		:Optional[List[str] | Tuple[str]]
	force_critical	:Optional[List[str] | Tuple[str]]
	contribfmt		:Optional[Dict[str,str]]
	CONTRIB_HOOK	=True


	def __init__(
					self,
					*,
					handler			:Optional[str]						=None,
					init_name		:Optional[str]						=None,
					init_level		:Optional[int]						=None,
					force_handover	:Optional[bool]						=None,
					force_debug		:Optional[List[str] | Tuple[str]]	=None,
					force_info		:Optional[List[str] | Tuple[str]]	=None,
					force_warning	:Optional[List[str] | Tuple[str]]	=None,
					force_critical	:Optional[List[str] | Tuple[str]]	=None,
					force_error		:Optional[List[str] | Tuple[str]]	=None,
					contribfmt		:Optional[Dict[str,str]]			=None
				):


		match (target := handler if handler is not None else getattr(self, "handler", stdout)):

			case str():
				if	ospath.isabs(target): makedirs(ospath.dirname(target), exist_ok=True)

				self.handler = FileHandler(target)
			case TextIOWrapper():
				self.handler = StreamHandler(handler)

			case _: raise TypeError("LibraryContrib handler must be file path string or stdout (default)")




		# It is allowed to provide "init_level" as any value that might be converted to integer.
		# Double float-int conversion assures almost any value (except None) that might be converted
		# to integer (floatable strings must be first converted to float) will be accepted.
		try:

			self.init_level = int(

				float(

					init_level if init_level is not None else
					getattr(self, "init_level", 20)
				)
			)

		except:	raise TypeError("LibraryContrib levels must be numeric values")




		self.init_name = str(

			init_name if init_name is not None else
			getattr(self, "init_name", "fantastic logs and where to contribute them")
		)


		self.contributor_name	= None
		self.contributor		= getLogger(self.init_name)
		self.handover_level		= self.init_level
		self.handover_map		= dict()
		self.handover_order		= list()
		self.handover_off		= True
		self.handover_name		= None


		self._handover_mode = bool(

			force_handover if force_handover is not None else
			getattr(self, "force_handover", False)
		)




		if		contribfmt is not None : fmtcandidate = contribfmt
		elif	getattr(self, "contribfmt", None) is not None : fmtcandidate = self.contribfmt
		else:	fmtcandidate = {

				"fmt": "%(asctime)s @%(name)s{origin} %(levelname)s : %(message)s",
				"datefmt": "%d/%m/%Y %H%M",
			}
		if(
			isinstance(fmtcandidate, dict)
			and
			fmtcandidate.get("fmt") is not None
			and
			isinstance(fmtcandidate.get("fmt"), str)
			and
			"{origin}" in fmtcandidate.get("fmt")
			and
			fmtcandidate.get("datefmt") is not None
			and
			isinstance(fmtcandidate.get("datefmt"), str)
		):
			self.contribfmt = fmtcandidate
		else:
			raise TypeError("No valid contributor formatter provided")


		self.makefmt("")
		self.contributor.addHandler(self.handler)
		self.contributor.setLevel(self.init_level)




		self.forcing_map = dict()
		self.forcing_map.update(
			{
				re.compile(P.replace("*",".*")): 10 for P in
				(
					force_debug if force_debug is not None and isinstance(force_debug, list | tuple)
					else
					getattr(self, "force_debug", None)
					if isinstance(getattr(self, "force_debug", None), list | tuple)
					else
					[]
				)
			}
		)
		self.forcing_map.update(
			{
				re.compile(P.replace("*",".*")): 20 for P in
				(
					force_info if force_info is not None and isinstance(force_info, list | tuple)
					else
					getattr(self, "force_info", None)
					if isinstance(getattr(self, "force_info", None), list | tuple)
					else
					[]
				)
			}
		)
		self.forcing_map.update(
			{
				re.compile(P.replace("*",".*")): 30 for P in
				(
					force_warning if force_warning is not None and isinstance(force_warning, list | tuple)
					else
					getattr(self, "force_warning", None)
					if isinstance(getattr(self, "force_warning", None), list | tuple)
					else
					[]
				)
			}
		)
		self.forcing_map.update(
			{
				re.compile(P.replace("*",".*")): 40 for P in
				(
					force_error if force_error is not None and isinstance(force_error, list | tuple)
					else
					getattr(self, "force_error", None)
					if isinstance(getattr(self, "force_error", None), list | tuple)
					else
					[]
				)
			}
		)
		self.forcing_map.update(
			{
				re.compile(P.replace("*",".*")): 50 for P in
				(
					force_critical if force_critical is not None and isinstance(force_critical, list | tuple)
					else
					getattr(self, "force_critical", None)
					if isinstance(getattr(self, "force_critical", None), list | tuple)
					else
					[]
				)
			}
		)








	def close(self): self.handler.close()
	def debug(self, message :str):
		if	self.handover_level <20:

			self.check_format(self.handover_name, self.handover_level)
			return self.contributor.debug(message)


	def info(self, message :str):
		if	self.handover_level <30:

			self.check_format(self.handover_name, self.handover_level)
			return self.contributor.info(message)


	def warning(self, message :str):
		if	self.handover_level <40:

			self.check_format(self.handover_name, self.handover_level)
			return self.contributor.warning(message)


	def error(self, message :str):
		if	self.handover_level <50:

			self.check_format(self.handover_name, self.handover_level)
			return self.contributor.error(message)


	def critical(self, message :str):
		if	self.handover_level <60:

			self.check_format(self.handover_name, self.handover_level)
			return self.contributor.critical(message)


	def log(self, level :int, message :str):

		""" Logging "log" imitation. """

		match level:
			case 10:	self.debug(message)
			case 20:	self.info(message)
			case 30:	self.warning(message)
			case 40:	self.error(message)
			case 50:	self.critical(message)
			case int():	self.contributor.log(level, message)








	def force(self, *layers :Tuple[str,...], level :int) -> int | TypeError :

		"""
			Helper method that allows altering levels for contributors.
			Accepts strings that must represent patterns that will "fullmatch" with corresponding callers
			representations in "handover_map". Any "layer" string not in "handover_map", or a "layer" with
			type that not string, will be just skipped. For every "layer" found sets corresponding level
			value to new "level", provided by mandatory keyword argument. If no "layers" provided, sets
			"level" for all objects in "handover_map".
			Maintains a counter of a number of objects which level mapping was changed and returns this
			counter. Raises TypeError in case "level" cannot be converted to int.
		"""

		try:	new_level = int(float(level))
		except:	raise TypeError(f"Level to force must be numeric, got \"{level}\"")
		else:

			counter = 0

			if	not layers:

				for name in self.handover_map : self.handover_map[name],counter = new_level,counter +1
			else:
				for layer in layers:

					if		isinstance(layer, str): current = re.compile(layer.replace("*",".*"))
					else:	continue


					for name in self.handover_map:
						if	current.fullmatch(name):

							self.handover_map[name] = new_level
							counter += 1


			return	counter


	def forced_or_init(self, name :str) -> int :

		"""
			Helper method that serves as logging level dispatcher according to "forcing_map".
			It is designed to obtain either forced or initial level for any object, that represented
			by "name" string. For such object "forcing_map" will be iterated and for very first key
			that caused "fullmatch" with "name" corresponding level will be returned. If no matches
			occurs, initial logging level will be returned. Raises TypeError if matched value is not
			an integer.
		"""

		for pattern,level in self.forcing_map.items():
			if	isinstance(pattern, re.Pattern):


				if	pattern.fullmatch(name):
					if	isinstance(level, int): return level


					raise	TypeError(f"Pattern \"{pattern.pattern}\" mapped to force incorrect \"{level}\"")
		else:		return	self.init_level








	def check_format(self, name :str, level :int, changeover=0):

		"""
			Utility method that estimates the "changeover" state for the current contribution.
			The keyword argument "changeover" will be xor'ed with corresponding bit that correspond
			to a certain state. The final value will be passed to a "bit_filter" method that must
			decide the following action for state.
			The "changeover state" comprises:
				- whether or not logging level was changed to a debug mode;
				- whether or not logging level was changed from debug mode;
				- whether or not logging object name was changed;
				- whether or handover mode is on;
				- whether or handover mode was turned on/off;
			In case of performance it is assumed that "name" and "level" are validated str and int.
			Returns None, handles no Exceptions.
		"""

		if	self.contributor.level != level:

			if		self.contributor.level <20 <= level : changeover ^= 1	# debug mode off->on (escalation)
			elif	level <20 <= self.contributor.level : changeover ^= 2	# debug mode off->on (decrease)

			self.contributor.setLevel(level)
		if	self.contributor.level == 10 :	changeover ^= 8
		if	self._handover_mode	:			changeover ^= 16
		if	self.handover_off	:			changeover ^= 32
		if	self.contributor_name != name:
			self.contributor_name = name
			changeover ^= 4


		self.bit_filter(changeover)




	def bit_filter(self, mask :int):

		"""
			Utility method that puts "mask" integer, that must represent "changeover state" for contribution,
			to a bitwise filter, with certain mask's values mapped with corresponding actions to contributor.
			This algorithm is the way to ensure proper formatting for any object, that contributes a log,
			along with minimization of such reformatting. Returns None, doesn't handle the only source of
			Exceptions from "makefmt" method.

			Changeover state mask values table for bitwise filtering:

			62 (32 + 16 + 8 + 4 + 2)	- handover off->on, debug mode off->on, name changed
			60 (32 + 16 + 8 + 4)		- handover off->on and name changed in debug mode
			58 (32 + 16 + 8 + 2)		- handover off->on, debug mode off->on
			56 (32 + 16 + 8)			- handover off->on in debug mode
			53 (32 + 16 + 4 + 1)		- handover off->on, debug mode on->off, name changed
			52 (32 + 16 + 4)			- handover off->on, name changed
			48 (32 + 16)				- handover off->on
			46 (32 + 8 + 4 + 2)			- debug mode off->on, name changed
			44 (32 + 8 + 4)				- name changed in debug mode
			42 (32 + 8 + 2)				- debug mode off->on
			37 (32 + 4 + 1)				- debug mode on->off, name changed
			33 (32 + 1)					- debug mode on->off
			30 (16 + 8 + 4 + 2)			- handover mode, debug mode off->on, name changed
			28 (16 + 8 + 4)				- handover mode, debug mode, name changed
			21 (16 + 4 + 1)				- handover mode, debug mode on->off, name changed
			20 (16 + 4)					- handover mode, name changed
			14 (8 + 4 + 2)				- handover mode on->off, debug mode off->on, name changed
			12 (8 + 4)					- handover mode on->off, debug mode, name changed
			10 (8 + 2)					- handover mode on->off, debug mode off->on
			8		 					- handover mode on->off, debug mode
			5 (4 + 1)					- handover mode on->off, debug mode on->off, name changed
			4							- handover mode on->off, name changed
			1							- handover mode on->off, debug mode on->off
			0							- handover mode on->off
		"""

		match mask:

			case 56:								self.handover_off = False
			case 10 | 8:							self.handover_off = True
			case 33 | 37:							self.makefmt("")
			case 46 | 44 | 42 | 30 | 28 | 21 | 20:	self.makefmt(f"-{self.contributor_name}")
			case 62 | 60 | 58 | 53 | 52 | 48 :

				self.makefmt(f"-{self.contributor_name}")
				self.handover_off = False

			case 14 | 12:

				self.makefmt(f"-{self.contributor_name}")
				self.handover_off = True

			case 5 | 4 | 1 | 0:

				self.makefmt("")
				self.handover_off = True




	def makefmt(self, handover_name :str) -> TypeError | ValueError | AttributeError :

		"""
			Utility method that does actual reformatting for contributor. As originally designed, there are
			only two forms of formatting for contributors:
				- general;
				- handover.
			By "handover formatting" current contributing object string representation is meant, which is
			included in the formatting string as {origin} variable. This case implies representation
			string for contributing object to be as close to object name as possible, so it was designed
			for a Transmutable class, that must returns it's full qualname as a string.
			In turn, "general formatting" omits contributing object name and makes format clear.
			It is very important, as current method relies on use of "contribfmt" dictionary as a
			pattern for reformatting, once it is subject of modification, the {origin} variable must
			be included in format string to ensure handover functionality (or such functionality must
			be ignored).
			The reformatting is handled directly on contributing "handler" object, which must be
			StreamHandler/FileHandler object. In case "contribfmt" is not a dictionary for proper
			formatting, 
			Returns None, raises Exception in case "handler"/"contribfmt" fields doesn't point
			to a valid object.
		"""

		if	hasattr(self, "handler"):
			if	callable(getattr(self.handler, "setFormatter", None)):
				if	isinstance(self.contribfmt, dict) and isinstance(self.contribfmt.get("fmt"), str):


					current_fmt = self.contribfmt.copy()
					current_fmt["fmt"] = current_fmt["fmt"].format(origin=handover_name)
					self.handler.setFormatter(Formatter(**current_fmt))


				else:	raise TypeError(f"Invalid contributor formatting dictionary \"{self.contribfmt}\"")
			else:		raise ValueError(f"Invalid contributor handler \"{self.handler}\"")
		else:			raise AttributeError("Contributor has no handler")








	def handover(self, caller :object, assign=True) -> LibraryContrib | ValueError :

		"""
			Core method that creates meta layer of maintaining contributor access.
			At the first place the "caller" object string representation is obtained. This string is
			used as a key to find corresponding logging level in "handover_map". Those two values
			will be then assigned as corresponding fields "handover_name" and "handover_level", which
			might be used then to maintain logging formatting. Additional the loggy assignment to
			a caller which level mapping is None, so it is it's first handover, is regulated by a boolean
			flag "assign". As Transmutables assign is automated, this regulation allows any other "caller"
			that not support attribute assignments to be maintained in "handover_map".
			It is assumed, that for correct formatting "handover" must be called for every object before
			such object will get access to contributing, as originally designed in Transmutable class that
			"handover" is called in every __getattribue__ method.
			As this functionality brings nothing but fancy logging for some stringable objects, it is
			might be skipped with no regrets to use contributing functionality only and directly.
			Returns itself as a pass after succesful "handover", or raises ValueError in only case,
			when "handover_map" has not an integer mapping for a "caller" string - in this case every
			call for "handover" for such "caller" will raise ValueError, even when try to remap value
			in "handover_map", so it is very important to not mess with handover mapping.
		"""


		name	= str(caller)
		level	= self.handover_map.get(name)


		match level:


			# Processing current level.
			# "None" must mean current "caller" has no mapping and hence was not assigned a contributor
			# yet, so it must go through the procedure and the level will be adjusted anyway.
			# "level" is integer means current "caller" was assigned it's contributor and mapped, so
			# the only thing need to check whether current level is differ from mapped.
			# Any other value for level means incorrect mapping and ValueError will be raised.
			case None:


				self.handover_map[name] = level = self.forced_or_init(name)
				self.handover_order.append(name)


				if	assign:
					try:	setattr(caller, "loggy", self)
					except:	pass


			case int():	pass
			case _:		raise ValueError(f"Incorrect handover mapping \"{name}\": \"{level}\"")


		self.handover_name	= name
		self.handover_level	= level


		return	self


	@property
	def handover_mode(self): return bool(self._handover_mode)
	@handover_mode.setter
	def handover_mode(self, state :bool): self._handover_mode = bool(state)
	def handovers(self, level :Optional[int] =None):

		"""
			Helper method that allows to obtain "handover_order" itself - a list of the strings, that
			represents the order in which handover was proceeded. An optional keyword argument "level"
			allows to provide an integer level which will be used as an announcer of "handover_order"
			list. Despite the level, "handover_mode" will be forced automatically to ensure informative
			logging, and by the end will be set to the state it was before the announce.
			Fields "handover_name" and "handover_level" will be set, imitating "handover" call.
			Then "log" method will be used to imitate regular logging.
		"""

		if	isinstance(level, int):

			self.handover_name	= "_handovers_"
			self.handover_level	= level
			handover_state		= self._handover_mode
			self._handover_mode	= True


			for name in self.handover_order : self.log(level, name)
			self._handover_mode = handover_state


		return	self.handover_order[::]










from time												import time
from time												import sleep
from multiprocessing									import Process
from typing												import Any
from typing												import Callable
from pygwarts.magical.philosophers_stone				import Transmutable
from pygwarts.magical.philosophers_stone.transmutations	import Transmutation
from pygwarts.magical.philosophers_stone.transmutations	import ControlledTransmutation
from pygwarts.magical.spells							import geminio
from pygwarts.magical.spells							import flagrate








class Callstamp(Transmutation):

	"""
		Utility decorator, that allows timestamping __call__ invocation of the decorated Transmutable
		object. Creates 2 time.time points, one before and one after the decorated object invocation,
		calculates the difference in seconds between them and emits corresponding info message. The
		result of decorated object __call__ will be the value to return. If decorated object doesn't
		implement __call__, raises corresponding TypeError, the way any non callable should.
	"""

	def _mutable_chain_injection(self, layer :Transmutable) -> Transmutable :
		class Callstamped(geminio(layer)):

			def __call__(self, *args, **kwargs) -> Any :

				if	hasattr(super(), "__call__"):

					start_point	= time()
					self.loggy.debug(f"Start point {start_point} created")


					output		= super().__call__(*args, **kwargs)
					end_point	= time()


					self.loggy.debug(f"End point {end_point} created")
					self.loggy.info(f"{self} finished in {end_point - start_point} seconds")


					return	output


				raise TypeError(f"{self} object is not callable")
		return	Callstamped








def mostsec(value :int | float | str, positive=False) -> str | None :

	"""
		Utility function that converts numerical value into a concise informative string representation
		of a time period in seconds, which that numerical value represents. Accepts any floatable value,
		such as int, float, exponent and their string representations. If "value" is negative, the absolute
		value will be taken. As "value" will be mandatory converted to float beforehand, any Exception raise
		will mean invalid "value" and "mostsec" will silently return None. The logic behind actual
		converting is inferring of at most two biggest measures for the "value", so the bigger "value" the
		more it will be truncated. There are upper "day" and lower "nano second" bounds. As conversion
		logic doesn't suggest 0 as a value, "mostsec" arithmetic considers 0 to be less than 1 nano second.
		The boolean flag "positive" toggles which value will be returned for a 0 "value" conversion. If
		positive=False (by default) the return value is "<1 ns", and None otherwise. This rule refers to
		such constraint as values less than 1 nano second are considered infinitesimals, so they might
		be approximated to 0, so positive=True will refer to negation of not positive values use.

		Comprehensive description for various values conversions:

			31539600	-> 365 d 1 h
			31539599	-> 365 d
			31536000	-> 365 d
			31535999	-> 364 d 23 h
			90000		-> 1 d 1 h
			89999		-> 1 d
			86400		-> 1 d
			86399		-> 23 h 59 m
			3660		-> 1 h 1 m
			3659		-> 1 h
			3600		-> 1 h
			3599		-> 59 m 59 s
			119			-> 1 m 59 s
			60 			-> 1 m
			59 			-> 59 s
			.9999		-> 999 ms
			.1999		-> 199 ms
			.1000		-> 100 ms
			.0999		-> 99 ms
			.0010		-> 1 ms
			.0009999	-> 999 us
			.0001000	-> 100 us
			.0000999	-> 99 us
			.00000100	-> 1 us
			.0000009999	-> 999 ns
			.0000001000	-> 100 ns
			.0000000999	-> 99 ns
			.0000000010	-> 1 ns
			.0000000001	-> <1 ns
			0			-> <1 ns
	"""

	try:	TIME = abs(float(value))
	except:	return
	else:

		if	not positive or (positive and TIME):

			mm,SS = divmod(TIME, 60)
			hh,MM = map(int, divmod(mm, 60))
			DD,HH = map(int, divmod(hh, 24))


			if	mm <1 :
				if	SS <1 :
					if	SS*1_000 <1 :
						if	SS*1_000_000 <1 :
							if	SS*1_000_000_000 <1 :


								return	"<1 ns"
							return	f"{int(SS*1_000_000_000)} ns"
						return	f"{int(SS*1_000_000)} us"
					return	f"{int(SS*1_000)} ms"
				return	f"{int(SS)} s"


			if	DD <1 :
				if	hh <1 :


					return	f"{MM} m" if SS <1 else f"{MM} m {int(SS)} s"
				return	f"{HH} h" if MM <1 else f"{HH} h {MM} m"
			return	f"{DD} d" if HH <1 else f"{DD} d {HH} h"








class mostsecfmt(ControlledTransmutation):

	"""
		Utility decorator that integrates "mostsec" function into mutable chain. Operates only with
		__call__ method of decorated Transmutable object. Raises TypeError in case __call__ not
		implemented for a decorated layer, the way any non callable should. As this decorator
		acts as numeric value interceptor, which must convert that value in some representing
		string, any invalid values will result None return from "mostsec", which provides conversion.
	"""

	def __init__(self, *, positive :bool): self.positive = positive
	def _mutable_chain_injection(self, layer :Transmutable) -> Transmutable :

		positive = self.positive


		class MostSec(geminio(layer)):
			def __call__(self, *args, **kwargs) -> str | None :

				if	hasattr(super(), "__call__"):
					if	(seconds := super().__call__(*args, **kwargs)) is not None :


						self.loggy.debug(f"Received \"{seconds}\" second{flagrate(seconds)} to format")
						return	mostsec(seconds, positive)


					else:	self.loggy.debug("Seconds to format was not received")
				else:		raise TypeError(f"{self} object is not callable")


		return	MostSec








class DIRTtimer(ControlledTransmutation):

	"""
		Utility decorator that serves as a time dispatcher for a decorated Transmutable object invocation.
		Accepts following numeric only key-word arguments:
			D		- time in seconds for delay before decorated Transmutable object invocation;
			I		- time in seconds for delay after decorated Transmutable object invocation;
			R		- counter for how many times decorated Transmutable object will be invoked;
			T		- time in seconds to terminate decorated Transmutable object invocation;
			spawner	- callable object to spawn process manageable for termination.
		Four numeric values must be convertible to float, and "T","D" and "I" will remain float, cause they
		represents time in seconds. The "R" value will be additionally converted to int, cause it will be
		used as iteration counter. This decorator relies on decorated Transmutable object implements
		__call__, which will be the target for dispatching, otherwise TypeError will be raised. Basically
		all the "DIRTtimer" does it's calling decorated __call__ with delay "D" and timer "T", with post
		delay "I", and repeats it "R" times. Delays "D" and "I" are just time.sleep calls, repetition of
		"R" times is just a for loop. The most crucial thing is the implementation of termination with
		timer "T" by "spawner". If T=0 (by default) decorated __call__ will be just invoked, according to
		"D","R","I" options. For any number of seconds, other than 0, "spawner" must create a parallel
		object, which will invoke decorated __call__ and stop it when the timer "T", that is time.sleep
		invocation, will be gone. By default "spawner" utilizes "multiprocessing" standard library module,
		which "Process" object work depends on running platform. For POSIX "Process.start" will use "fork"
		context and this is the originally designed and tested feature. For NT "Process.start" uses "spawn"
		context which causes serialization error by "pickle" module, involved in "multiprocessing". In that
		case, "spawner" should be set to any other object, capable of handling "spawn" context. With T=0
		no new spawns occur, so decorated Transmutable object persist through all "D","R","I" manipulations.
		For any 0<T every new spawn means new decorated object, so no modification can persist via "R"
		repetitions. Also, as decorated __call__ repeatedly invocation is not a subject for results
		collecting, there is no return value, so this must be handled in decorated Transmutable object.

		--------------------------------------------------------------------------------------------------
		| This class designed as a very simple and concise way of objects calling manipulation, so the	 |
		| termination part ain't complies modern best practices, and since python3.12 "fork()" seems to	 |
		| be deprecated. The use of such functionality (termination) must be the subject to consider,	 |
		| unless the other solution becomes the part of pygwarts code base.								 |
		--------------------------------------------------------------------------------------------------
	"""

	def __init__(
					self,
					*,
					D		:int | float | str		=0,			# Delay timer
					I		:int | float | str		=0,			# Delay between intervals, also post delay
					R		:int | float | str		=1,			# Repetition counter
					T		:int | float | str		=0,			# Termination timer
					spawner	:Callable[[Any],Any]	=Process	# Object to spawn a terminatable process
				):


		try:	self.delay_sleep = abs(float(D))
		except:	raise ValueError(f"Delay timer \"{D}\" is invalid")


		try:	self.interval_sleep = abs(float(I))
		except:	raise ValueError(f"Interval timer \"{I}\" is invalid")


		try:	self.repeat_times = abs(int(float(R)))
		except:	raise ValueError(f"Repetition counter \"{R}\" is invalid")


		try:	self.terminate_sleep = abs(float(T))
		except:	raise ValueError(f"Termination timer \"{T}\" is invalid")


		if(
				callable(spawner) and
				callable(getattr(spawner, "start", None)) and
				callable(getattr(spawner, "terminate", None))
		):		self.spawner = spawner
		else:	raise ValueError(f"Process spawner \"{spawner}\" is invalid")


	def _mutable_chain_injection(self, layer :Transmutable) -> Transmutable :


		D = self.delay_sleep
		I = self.interval_sleep
		R = self.repeat_times
		T = self.terminate_sleep
		spawner = self.spawner


		class DIRT(geminio(layer)):
			def __call__(self, *args, **kwargs):


				if	hasattr(super(), "__call__"):


					self.loggy.debug(f"Delay start timer {D} second{flagrate(D)}")
					self.loggy.debug(f"Interval delay {I} second{flagrate(I)}")
					self.loggy.debug(f"Repetition counter {R} time{flagrate(R)}")
					self.loggy.debug(f"Termination timer {T} second{flagrate(T)}")
					self.loggy.debug(f"Caller arguments: {args}")
					self.loggy.debug(f"Caller keyword arguments: {kwargs}")


					for i in range(R):
						self.loggy.debug(f"DIRT iteration {i +1}")


						if	D:
							sleep(D)
							self.loggy.debug(f"Delay {D} second{flagrate(D)} for {self} performed")


						if	T:
							self.loggy.info(f"Starting {T} second{flagrate(T)} timer for {self}")


							process = spawner(target=super().__call__, args=args, kwargs=kwargs,)
							process.start()
							sleep(T)
							process.terminate()


							self.loggy.debug(f"Process {self} terminated")
						else:
							super().__call__(*args, **kwargs)


						if	I:
							sleep(I)
							self.loggy.debug(f"Interval {I} second{flagrate(I)} for {self} performed")


				else:	raise TypeError(f"{self} object is not callable")


		return	DIRT







